/********************************************************************************/
/* lorenz.c: Definition of Functions and Constants in Lorenz Model              */
/* Copyright (C) 2011 Tomonori Kouya                                            */
/*                                                                              */
/* This program is free software: you can redistribute it and/or modify it      */
/* under the terms of the GNU Lesser General Public License as published by the */
/* Free Software Foundation, either version 3 of the License or any later       */
/* version.                                                                     */
/*                                                                              */
/* This program is distributed in the hope that it will be useful, but WITHOUT  */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or        */
/* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License */
/* for more details.                                                            */
/*                                                                              */
/* You should have received a copy of the GNU Lesser General Public License     */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                              */
/********************************************************************************/
/* double */
#define SIGMA (10.0)
#define R (470.0 / 19.0)
#define B (8.0 / 3.0)

void ddf(DVector y, double x0, DVector y0)
{
	/* y'[0] = -sigma * y[0] + sigma * y[1]   */
	/* y'[1] = -y[0] * y[2] + r * y[0] - y[1] */
	/* y'[2] = y[0] * y[1] - b * y[2]         */
	set_dvector_i(y, 0, -SIGMA * gdvi(y0, 0) + SIGMA * gdvi(y0, 1));
	set_dvector_i(y, 1, -gdvi(y0, 0) * gdvi(y0, 2) + R * gdvi(y0, 0) - gdvi(y0, 1));
	set_dvector_i(y, 2, gdvi(y0, 0) * gdvi(y0, 1) - B * gdvi(y0, 2));

	return;
}

void ddfj(DMatrix ymat, double x0, DVector y0)
{
	/* ymat = [   -sigma sigma    0  ] */
	/*        [ r - y[2]    -1 -y[0] ] */
	/*        [     y[1]  y[0]   -b  ] */
	set_dmatrix_ij(ymat, 0, 0, -SIGMA);
	set_dmatrix_ij(ymat, 0, 1, SIGMA);
	set_dmatrix_ij(ymat, 0, 2, 0.0);
	set_dmatrix_ij(ymat, 1, 0, R - gdvi(y0, 2));
	set_dmatrix_ij(ymat, 1, 1, -1.0);
	set_dmatrix_ij(ymat, 1, 2, -gdvi(y0, 0));
	set_dmatrix_ij(ymat, 2, 0, gdvi(y0, 1));
	set_dmatrix_ij(ymat, 2, 1, gdvi(y0, 0));
	set_dmatrix_ij(ymat, 2, 2, -B);

	return;
}

/* y = [exp(x), ...., exp(x)]^T */
void dans(DVector y, double x)
{
	long int i;

	for(i = 0; i < y->dim; i++)
		set_dvector_i(y, i, exp(x));

	return;
}

#ifdef USE_GMP

#define MPF_SIGMA "10"
#define MPF_R_NUM "470"
#define MPF_R_DEN "19" // R = R_NUM / R_DEN
#define MPF_B_NUM "8"
#define MPF_B_DEN "3" // B = B_NUM / B_DEN

/* mpf_t */
void df(MPFVector y, mpf_t x0, MPFVector y0)
{
	unsigned long prec;
	mpf_t sigma, r, b, tmp[3];

	prec = y->prec;

	mpf_init2(sigma, prec);
	mpf_init2(r, prec);
	mpf_init2(b, prec);
	mpf_init2(tmp[0], prec);
	mpf_init2(tmp[1], prec);
	mpf_init2(tmp[2], prec);

	mpf_set_str(sigma, MPF_SIGMA, 10);
	mpf_set_str(tmp[0], MPF_R_NUM, 10);
	mpf_set_str(tmp[1], MPF_R_DEN, 10);
	mpf_div(r, tmp[0], tmp[1]);
	mpf_set_str(tmp[0], MPF_B_NUM, 10);
	mpf_set_str(tmp[1], MPF_B_DEN, 10);
	mpf_div(b, tmp[0], tmp[1]);

	/* y'[0] = -sigma * y[0] + sigma * y[1]   */
	mpf_mul(tmp[0], sigma, gmpfvi(y0, 0));
	mpf_neg(tmp[0], tmp[0]);
	mpf_mul(tmp[1], sigma, gmpfvi(y0, 1));
	mpf_add(tmp[2], tmp[0], tmp[1]);
	set_mpfvector_i(y, 0, tmp[2]);

	/* y'[1] = -y[0] * y[2] + r * y[0] - y[1] */
	mpf_mul(tmp[0], gmpfvi(y0, 0), gmpfvi(y0, 2));
	mpf_neg(tmp[0], tmp[0]);
	mpf_mul(tmp[1], r, gmpfvi(y0, 0));
	mpf_add(tmp[2], tmp[0], tmp[1]);
	mpf_sub(tmp[2], tmp[2], gmpfvi(y0, 1));
	set_mpfvector_i(y, 1, tmp[2]);

	/* y'[2] = y[0] * y[1] - b * y[2]         */
	mpf_mul(tmp[0], gmpfvi(y0, 0), gmpfvi(y0, 1));
	mpf_mul(tmp[1], b, gmpfvi(y0, 2));
	mpf_sub(tmp[2], tmp[0], tmp[1]);
	set_mpfvector_i(y, 2, tmp[2]);

	mpf_clear(sigma);
	mpf_clear(r);
	mpf_clear(b);
	mpf_clear(tmp[0]);
	mpf_clear(tmp[1]);
	mpf_clear(tmp[2]);

	return;
}

void dfj(MPFMatrix ymat, mpf_t x0, MPFVector y0)
{
	unsigned long prec;
	mpf_t sigma, r, b, tmp[2];

	prec = ymat->prec;

	mpf_init2(sigma, prec);
	mpf_init2(r, prec);
	mpf_init2(b, prec);
	mpf_init2(tmp[0], prec);
	mpf_init2(tmp[1], prec);

	mpf_set_str(sigma, MPF_SIGMA, 10);
	mpf_set_str(tmp[0], MPF_R_NUM, 10);
	mpf_set_str(tmp[1], MPF_R_DEN, 10);
	mpf_div(r, tmp[0], tmp[1]);
	mpf_set_str(tmp[0], MPF_B_NUM, 10);
	mpf_set_str(tmp[1], MPF_B_DEN, 10);
	mpf_div(b, tmp[0], tmp[1]);

	/* ymat = [   -sigma sigma    0  ] */
	/*        [ r - y[2]    -1 -y[0] ] */
	/*        [     y[1]  y[0]   -b  ] */
	mpf_neg(tmp[0], sigma);
	set_mpfmatrix_ij(ymat, 0, 0, tmp[0]);
	set_mpfmatrix_ij(ymat, 0, 1, sigma);
	mpf_set_ui(tmp[0], 0UL);
	set_mpfmatrix_ij(ymat, 0, 2, tmp[0]);
	mpf_sub(tmp[0], r, gmpfvi(y0, 2));
	set_mpfmatrix_ij(ymat, 1, 0, tmp[0]);
	mpf_set_ui(tmp[0], 1Ul); mpf_neg(tmp[0], tmp[0]);
	set_mpfmatrix_ij(ymat, 1, 1, tmp[0]);
	mpf_neg(tmp[0], gmpfvi(y0, 0));
	set_mpfmatrix_ij(ymat, 1, 2, tmp[0]);
	set_mpfmatrix_ij(ymat, 2, 0, gmpfvi(y0, 1));
	set_mpfmatrix_ij(ymat, 2, 1, gmpfvi(y0, 0));
	mpf_neg(tmp[0], b);
	set_mpfmatrix_ij(ymat, 2, 2, tmp[0]);

	mpf_clear(sigma);
	mpf_clear(r);
	mpf_clear(b);
	mpf_clear(tmp[0]);
	mpf_clear(tmp[1]);

	return;
}

/* y = [exp(x), ...., exp(x)]^T */
void ans(MPFVector y, mpf_t x)
{
	long int i;
	mpf_t tmp;

	mpf_init(tmp);
	for(i = 0; i < y->dim; i++)
	{
		mpf_exp(tmp, x);
		set_mpfvector_i(y, i, tmp);
	}
	mpf_clear(tmp);
	return;
}


void initial_value(mpf_t lf_initx, mpf_t lf_maxx, MPFVector lf_inity)
{
	long int i;

	/* x0 := 0 */
	mpf_init_set_str(lf_initx, "0.0", 10);
	mpf_init_set_str(lf_maxx, "50.0", 10);

	/* y := [0 1 0]^T */
	set_mpfvector_i_ui(lf_inity, 0, 0UL);
	set_mpfvector_i_ui(lf_inity, 1, 1UL);
	set_mpfvector_i_ui(lf_inity, 2, 0UL);

}
#endif

void initial_dvalue(double *lf_initx, double *lf_maxx, DVector lf_inity)
{
	long int i;

	/* x0 := 0 */
	*lf_initx = 0.0;
	*lf_maxx = 50.0;

	/* y := [0 1 0]^T */
	set_dvector_i(lf_inity, 0, 0.0);
	set_dvector_i(lf_inity, 1, 1.0);
	set_dvector_i(lf_inity, 2, 0.0);
}

#define DIM 3
