/********************************************************************************/
/* rkcoef.c : Coefficients of Explicit & Implicit Runge-Kutta Methods           */
/* Copyright (C) 2011 Tomonori Kouya                                            */
/*                                                                              */
/* This program is free software: you can redistribute it and/or modify it      */
/* under the terms of the GNU Lesser General Public License as published by the */
/* Free Software Foundation, either version 3 of the License or any later       */
/* version.                                                                     */
/*                                                                              */
/* This program is distributed in the hope that it will be useful, but WITHOUT  */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or        */
/* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License */
/* for more details.                                                            */
/*                                                                              */
/* You should have received a copy of the GNU Lesser General Public License     */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                              */
/********************************************************************************/
#include <stdio.h>
#include <math.h>
#include "bncode.h"

/* get coefficients for erk */
void get_derk_coef(unsigned long coef_type, DVector coef_c, DMatrix coef_a, DVector coef_w)
{
	switch(coef_type)
	{
		case ERK_OPTIMAL_2_2:
			
			/* c1 = 2/3, a11 = 2/3, w1 = 1/4, w2 = 3/4 */
			sdvi (coef_c, 0, (double)2 / 3);
			sdmij(coef_a, 0, 0, (double)2 / 3);
			sdvi (coef_w, 0, (double)1 / 4);
			sdvi (coef_w, 1, (double)3 / 4);
			
			break;

		case ERK_CLASSICAL_4_4:

			 /* c1 = 1/2, c2 = 1/2, c3 = 1 */
			sdvi(coef_c, 0, 0.5);
			sdvi(coef_c, 1, 0.5);
			sdvi(coef_c, 2, 1);

			/* w1 = 1/6, w2 = 2/6, w3 = 2/6, w4 = 1/6 */
			sdvi(coef_w, 0, (double)1 / 6);
			sdvi(coef_w, 1, (double)2 / 6);
			sdvi(coef_w, 2, (double)2 / 6);
			sdvi(coef_w, 3, (double)1 / 6);

			/* a11 = 1/2, a22 = 1/2, a33 = 1 */
			sdmij(coef_a, 0, 0, 0.5);
			sdmij(coef_a, 1, 1, 0.5);
			sdmij(coef_a, 2, 2, 1);
			
			break;

		case ERK_BUTCHER1_7_6:

			/* c1 = 1/3, c2 = 2/3, c3 = 1/3, c4 = 1/2, c5 = 1/2, c6 = 1 */
			/* c1 = 1/3, c2 = 2/3, c3 = 1/3, c4 = 1/2, c5 = 1/2, c6 = 1 */
			sdvi(coef_c, 0, (double)1 / 3);
			sdvi(coef_c, 1, (double)2 / 3);
			sdvi(coef_c, 2, (double)1 / 3);
			sdvi(coef_c, 3, 0.5);
			sdvi(coef_c, 4, 0.5);
			sdvi(coef_c, 5, 1);
			
			/* w1 = 11/120, w2 = 0, w3, = 27/40, w4 = 27/40, w5 = -4/15, w6 = -4/15, w7 = 11/120 */
			/* w1 = 11/120, w2 = 0, w3, = 27/40, w4 = 27/40, w5 = -4/15, w6 = -4/15, w7 = 11/120 */
			sdvi(coef_w, 0, (double)11 / 120);
			sdvi(coef_w, 2, (double)27 / 40);
			sdvi(coef_w, 3, (double)27 / 40);
			sdvi(coef_w, 4, (double)(-4) / 15);
			sdvi(coef_w, 5, (double)(-4) / 15);
			sdvi(coef_w, 6, (double)11 / 120);
			
			/* a11 = 1/3 */
			/* a21 = 0, a22 = 2/3 */
			/* a31 = 1/12, a32 = 1/3, a33 = -1/12 */
			/* a41 = -1/16, a42 = 9/8, a43 = -3/16, a44 = -3/8 */
			/* a51 = 0, a52 = 9/8, a53 = -3/8, a54 = -3/4, a55 = 1/2 */
			/* a61 = 9/44, a62 = -9/11, a63 = 63/44, a64 = 18/11, a65 = 0, a66 = -16/11 */
			/* a11 = 1/3 */
			/* a21 = 0, a22 = 2/3 */
			/* a31 = 1/12, a32 = 1/3, a33 = -1/12 */
			/* a41 = -1/16, a42 = 9/8, a43 = -3/16, a44 = -3/8 */
			/* a51 = 0, a52 = 9/8, a53 = -3/8, a54 = -3/4, a55 = 1/2 */
			/* a61 = 9/44, a62 = -9/11, a63 = 63/44, a64 = 18/11, a65 = 0, a66 = -16/11 */
			sdmij(coef_a, 0, 0, (double)1 / 3);
			sdmij(coef_a, 1, 1, (double)2 / 3);
			sdmij(coef_a, 2, 0, (double)1 / 12);
			sdmij(coef_a, 2, 1, (double)1 / 3);
			sdmij(coef_a, 2, 2, (double)(-1) / 12);
			sdmij(coef_a, 3, 0, (double)(-1) / 16);
			sdmij(coef_a, 3, 1, (double)9 / 8);
			sdmij(coef_a, 3, 2, (double)(-3) / 16);
			sdmij(coef_a, 3, 3, (double)(-3) / 8);
			sdmij(coef_a, 4, 1, (double)9 / 8);
			sdmij(coef_a, 4, 2, (double)(-3) / 8);
			sdmij(coef_a, 4, 3, (double)(-3) / 4);
			sdmij(coef_a, 4, 4, 0.5);
			sdmij(coef_a, 5, 0, (double)9 / 44);
			sdmij(coef_a, 5, 1, (double)(-9) / 11);
			sdmij(coef_a, 5, 2, (double)63 / 44);
			sdmij(coef_a, 5, 3, (double)18 / 11);
			sdmij(coef_a, 5, 5, (double)(-16) / 11);

			break;

		default:
			fprintf(stderr, "get_derk_coef: Such a coefficient set not exist.\n");
	}

}

/* set coefficients */
void get_dirk_coef(unsigned long coef_type, DVector coef_c, DMatrix coef_a, DVector coef_w)
{

	switch(coef_type)
	{

		case IRK_GAUSS_1_2:
			
			/* c_1 = 1/2, w_1 = 1, a_11 = 1/2 */
			sdvi(coef_c, 0, 0.5);
			sdvi(coef_w, 0, 1);
			sdmij(coef_a, 0, 0, 0.5);
			
			break;

		case IRK_GAUSS_2_4:
			
			/* c_1 = (3-3^(1/2))/6, c_2 = (3+3^(1/2))/6 */
			sdvi(coef_c, 0, ((double)3 - sqrt((double)3)) / 6);
			sdvi(coef_c, 1, ((double)3 + sqrt((double)3)) / 6);

			/* w1 = 1/2, w2 = 1/2 */
			sdvi(coef_w, 0, 0.5);
			sdvi(coef_w, 1, 0.5);

			/* a11 = 1/4, a12 = (3 - 2*3^(1/2))/12 */
			/* a12 = (3 + 2*3^(1/2))/12, a22 = 1/4 */
			sdmij(coef_a, 0, 0, 0.25);
			sdmij(coef_a, 0, 1, ((double)3 - 2 * sqrt((double)3)) / 12);
			sdmij(coef_a, 1, 0, ((double)3 + 2 * sqrt((double)3)) / 12);
			sdmij(coef_a, 1, 1, 0.25);
			
			break;

		case IRK_GAUSS_3_6:

			/* c1 = (5 - 15^(1/2))/10, c2 = 1/2, c3 = (5 + 15^(1/2))/10 */
			sdvi(coef_c, 0, ((double)5 - sqrt((double)15)) / 10);
			sdvi(coef_c, 1, 0.5);
			sdvi(coef_c, 2, ((double)5 + sqrt((double)15)) / 10);

			/* w1 = 5/18, w2 = 8/18, w3 = 5/18 */
			sdvi(coef_w, 0, (double)5 / 18);
			sdvi(coef_w, 1, (double)8 / 18);
			sdvi(coef_w, 2, (double)5 / 18);

			/* a11 = 5/36, a12 = (10-3*15^(1/2))/45, a13 = (25-6*15^(1/2))/180 */
			/* a21 = (10+3*15^(1/2))/72, a22 = 2/9, a23 = (10-3*15^(1/2))/72   */
			/* a31 = (25+6*15^(1/2))/180, a32 = (10+3*15^(1/5))/45, a33 = 5/36 */
			sdmij(coef_a, 0, 0, (double)5 / 36);
			sdmij(coef_a, 0, 1, ((double)10 - 3 * sqrt((double)15)) / 45);
			sdmij(coef_a, 0, 2, ((double)25 - 6 * sqrt((double)15)) / 180);
			sdmij(coef_a, 1, 0, ((double)10 + 3 * sqrt((double)15)) / 72);
			sdmij(coef_a, 1, 1, (double)2 / 9);
			sdmij(coef_a, 1, 2, ((double)10 - 3 * sqrt((double)15)) / 72);
			sdmij(coef_a, 2, 0, ((double)25 + 6 * sqrt((double)15)) / 180);
			sdmij(coef_a, 2, 1, ((double)10 + 3 * sqrt((double)15)) / 45);
			sdmij(coef_a, 2, 2, (double)5 / 36);

			break;

		default:
			fprintf(stderr, "get_dirk_coef: Such a coefficient set not exist.\n");
	}

}

#ifdef USE_GMP
/* get coefficients for erk */
void get_mpf_erk_coef(unsigned long coef_type, unsigned long prec, MPFVector coef_c, MPFMatrix coef_a, MPFVector coef_w)
{
	mpf_t numerator, denominator, coef, tmp;

	mpf_init2(numerator, prec);
	mpf_init2(denominator, prec);
	mpf_init2(coef, prec);
	mpf_init2(tmp, prec);

	switch(coef_type)
	{
		case ERK_OPTIMAL_2_2:
//			coef_c = init2_mpfvector(1, prec);
//			coef_w = init2_mpfvector(2, prec);
//			coef_a = init2_mpfmatrix(1, 1, prec);
			
			/* c1 = 2/3, a11 = 2/3, w1 = 1/4, w2 = 3/4 */
			mpf_set_ui(numerator, 2UL);
			mpf_set_ui(denominator, 3UL);
			mpf_div(coef, numerator, denominator);
			smpfvi (coef_c, 0, coef);
			smpfmij(coef_a, 0, 0, coef);
			mpf_set_ui(numerator, 1UL);
			mpf_set_ui(denominator, 4UL);
			mpf_div(coef, numerator, denominator);
			smpfvi (coef_w, 0, coef);
			mpf_set_ui(numerator, 3UL);
			mpf_div(coef, numerator, denominator);
			smpfvi (coef_w, 1, coef);
			
			break;

		case ERK_CLASSICAL_4_4:
//			coef_c = init2_mpfvector(3, prec);
//			coef_w = init2_mpfvector(4, prec);
//			coef_a = init2_mpfmatrix(3, 3, prec);

			 /* c1 = 1/2, c2 = 1/2, c3 = 1 */
			mpf_set_ui(numerator, 1UL);
			mpf_set_ui(denominator, 2UL);
			mpf_div(coef, numerator, denominator);
			smpfvi(coef_c, 0, coef);
			smpfvi(coef_c, 1, coef);
			smpfvi(coef_c, 2, numerator);

			/* w1 = 1/6, w2 = 2/6, w3 = 2/6, w4 = 1/6 */
			mpf_set_ui(numerator, 1UL);
			mpf_set_ui(denominator, 6UL);
			mpf_div(coef, numerator, denominator);
			smpfvi(coef_w, 0, coef);
			smpfvi(coef_w, 3, coef);
			mpf_set_ui(numerator, 2UL);
			mpf_div(coef, numerator, denominator);
			smpfvi(coef_w, 1, coef);
			smpfvi(coef_w, 2, coef);

			/* a11 = 1/2, a22 = 1/2, a33 = 1 */
			mpf_set_ui(numerator, 1UL);
			mpf_set_ui(denominator, 2UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 0, 0, coef);
			smpfmij(coef_a, 1, 1, coef);
			smpfmij(coef_a, 2, 2, numerator);
			
			break;

		case ERK_BUTCHER1_7_6:
//			coef_c = init2_mpfvector(6, prec);
//			coef_w = init2_mpfvector(7, prec);
//			coef_a = init2_mpfmatrix(6, 6, prec);

			/* c1 = 1/3, c2 = 2/3, c3 = 1/3, c4 = 1/2, c5 = 1/2, c6 = 1 */
			mpf_set_ui(numerator, 1UL);
			mpf_set_ui(denominator, 3UL);
			mpf_div(coef, numerator, denominator);
			smpfvi(coef_c, 0, coef);
			smpfvi(coef_c, 2, coef);
			mpf_set_ui(numerator, 2UL);
			mpf_div(coef, numerator, denominator);
			smpfvi(coef_c, 1, coef);
			mpf_set_ui(numerator, 1UL);
			mpf_set_ui(denominator, 2UL);
			mpf_div(coef, numerator, denominator);
			smpfvi(coef_c, 3, coef);
			smpfvi(coef_c, 4, coef);
			smpfvi(coef_c, 5, numerator);
			
			/* w1 = 11/120, w2 = 0, w3, = 27/40, w4 = 27/40, w5 = -4/15, w6 = -4/15, w7 = 11/120 */
			mpf_set_ui(numerator, 11UL);
			mpf_set_ui(denominator, 120UL);
			mpf_div(coef, numerator, denominator);
			smpfvi(coef_w, 0, coef);
			smpfvi(coef_w, 6, coef);
			mpf_set_ui(numerator, 27UL);
			mpf_set_ui(denominator, 40UL);
			mpf_div(coef, numerator, denominator);
			smpfvi(coef_w, 2, coef);
			smpfvi(coef_w, 3, coef);
			mpf_set_si(numerator, -4);
			mpf_set_ui(denominator, 15UL);
			mpf_div(coef, numerator, denominator);
			smpfvi(coef_w, 4, coef);
			smpfvi(coef_w, 5, coef);
			
			/* a11 = 1/3 */
			/* a21 = 0, a22 = 2/3 */
			/* a31 = 1/12, a32 = 1/3, a33 = -1/12 */
			/* a41 = -1/16, a42 = 9/8, a43 = -3/16, a44 = -3/8 */
			/* a51 = 0, a52 = 9/8, a53 = -3/8, a54 = -3/4, a55 = 1/2 */
			/* a61 = 9/44, a62 = -9/11, a63 = 63/44, a64 = 18/11, a65 = 0, a66 = -16/11 */
			mpf_set_ui(numerator, 1UL);
			mpf_set_ui(denominator, 3UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 0, 0, coef);
			mpf_set_ui(numerator, 2UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 1, 1, coef);
			mpf_set_ui(numerator, 1UL);
			mpf_set_ui(denominator, 12UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 2, 0, coef);
			mpf_neg(coef, coef);
			smpfmij(coef_a, 2, 2, coef);
			mpf_set_ui(numerator, 1UL);
			mpf_set_ui(denominator, 3UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 2, 1, coef);
			mpf_set_si(numerator, -1);
			mpf_set_ui(denominator, 16UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 3, 0, coef);
			mpf_set_ui(numerator, 18UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 3, 1, coef);
			mpf_set_si(numerator, -3);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 3, 2, coef);
			mpf_set_si(numerator, -6);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 3, 3, coef);
			mpf_set_ui(numerator, 9UL);
			mpf_set_ui(denominator, 8UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 4, 1, coef);
			mpf_set_si(numerator, -3);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 4, 2, coef);
			mpf_set_si(numerator, -6);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 4, 3, coef);
			mpf_set_ui(numerator, 4UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 4, 4, coef);
			mpf_set_ui(numerator, 9UL);
			mpf_set_ui(denominator, 44UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 5, 0, coef);
			mpf_set_si(numerator, -36);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 5, 1, coef);
			mpf_set_ui(numerator, 63);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 5, 2, coef);
			mpf_set_ui(numerator, 72UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 5, 3, coef);
			mpf_set_si(numerator, -64);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 5, 5, coef);
			break;

		default:
			fprintf(stderr, "set_coef: Such a coefficient set not exist.\n");
	}

	mpf_clear(numerator);
	mpf_clear(denominator);
	mpf_clear(coef);
	mpf_clear(tmp);

}

/* set coefficients */
void get_mpf_irk_coef(unsigned long coef_type, unsigned long prec, MPFVector coef_c, MPFMatrix coef_a, MPFVector coef_w)
{
	mpf_t numerator, denominator, coef, tmp;

	mpf_init2(numerator, prec);
	mpf_init2(denominator, prec);
	mpf_init2(coef, prec);
	mpf_init2(tmp, prec);

	switch(coef_type)
	{

		case IRK_GAUSS_1_2:
//			coef_c = init2_mpfvector(1, prec);
//			coef_w = init2_mpfvector(1, prec);
//			coef_a = init2_mpfmatrix(1, 1, prec);
			
			/* c_1 = 1/2, w_1 = 1, a_11 = 1/2 */
			mpf_set_ui(numerator, 1UL);
			mpf_set_ui(denominator, 2UL);
			mpf_div(coef, numerator, denominator);
			smpfvi(coef_c, 0, coef);
			smpfvi(coef_w, 0, numerator);
			smpfmij(coef_a, 0, 0, coef);
			
			break;

		case IRK_GAUSS_2_4:
//			coef_c = init2_mpfvector(2, prec);
//			coef_w = init2_mpfvector(2, prec);
//			coef_a = init2_mpfmatrix(2, 2, prec);
			
			/* c_1 = (3-3^(1/2))/6, c_2 = (3+3^(1/2))/6 */
			mpf_set_ui(numerator, 3UL);
			mpf_sqrt_ui(tmp, 3UL);
			mpf_sub(numerator, numerator, tmp);
			mpf_set_ui(denominator, 6UL);
			mpf_div(coef, numerator, denominator);
			smpfvi(coef_c, 0, coef);
			mpf_set_ui(numerator, 3UL);
			mpf_add(numerator, numerator, tmp);
			mpf_div(coef, numerator, denominator);
			smpfvi(coef_c, 1, coef);

			/* w1 = 1/2, w2 = 1/2 */
			mpf_set_ui(numerator, 1UL);
			mpf_set_ui(denominator, 2UL);
			mpf_div(coef, numerator, denominator);
			smpfvi(coef_w, 0, coef);
			smpfvi(coef_w, 1, coef);

			/* a11 = 1/4, a12 = (3 - 2*3^(1/2))/12 */
			/* a12 = (3 + 2*3^(1/2))/12, a22 = 1/4 */
			mpf_set_ui(numerator, 1UL);
			mpf_set_ui(denominator, 4UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 0, 0, coef);
			smpfmij(coef_a, 1, 1, coef);
			mpf_set_ui(numerator, 3UL);
			mpf_sqrt_ui(tmp, 3UL); mpf_mul_ui(tmp, tmp, 2UL);
			mpf_sub(numerator, numerator, tmp);
			mpf_set_ui(denominator, 12UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 0, 1, coef);
			mpf_set_ui(numerator, 3UL);
			mpf_add(numerator, numerator, tmp);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 1, 0, coef);
			
			break;

		case IRK_GAUSS_3_6:
//			coef_c = init2_mpfvector(3, prec);
//			coef_w = init2_mpfvector(3, prec);
//			coef_a = init2_mpfmatrix(3, 3, prec);

			/* c1 = (5 - 15^(1/2))/10, c2 = 1/2, c3 = (5 + 15^(1/2))/10 */
			mpf_set_ui(numerator, 5UL);
			mpf_sqrt_ui(tmp, 15UL);
			mpf_sub(numerator, numerator, tmp);
			mpf_set_ui(denominator, 10UL);
			mpf_div(coef, numerator, denominator);
			smpfvi(coef_c, 0, coef);
			mpf_set_ui(numerator, 5UL);
			mpf_add(numerator, numerator, tmp);
			mpf_div(coef, numerator, denominator);
			smpfvi(coef_c, 2, coef);
			mpf_set_ui(numerator, 1UL);
			mpf_set_ui(denominator, 2UL);
			mpf_div(coef, numerator, denominator);
			smpfvi(coef_c, 1, coef);

			/* w1 = 5/18, w2 = 8/18, w3 = 5/18 */
			mpf_set_ui(numerator, 5UL);
			mpf_set_ui(denominator, 18UL);
			mpf_div(coef, numerator, denominator);
			smpfvi(coef_w, 0, coef);
			smpfvi(coef_w, 2, coef);
			mpf_set_ui(numerator, 8UL);
			mpf_div(coef, numerator, denominator);
			smpfvi(coef_w, 1, coef);

			/* a11 = 5/36, a12 = (10-3*15^(1/2))/45, a13 = (25-6*15^(1/2))/180 */
			/* a21 = (10+3*15^(1/2))/72, a22 = 2/9, a23 = (10-3*15^(1/2))/72   */
			/* a31 = (25+6*15^(1/2))/180, a32 = (10+3*15^(1/5))/45, a33 = 5/36 */
			mpf_set_ui(numerator, 5UL);
			mpf_set_ui(denominator, 36UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 0, 0, coef);
			smpfmij(coef_a, 2, 2, coef);
			mpf_set_ui(numerator, 8UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 1, 1, coef);
			mpf_set_ui(numerator, 10UL);
			mpf_sqrt_ui(tmp, 15UL);
			mpf_mul_ui(tmp, tmp, 3UL);
			mpf_sub(numerator, numerator, tmp);
			mpf_set_ui(denominator, 45UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 0, 1, coef);
			mpf_set_ui(denominator, 72UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 1, 2, coef);
			mpf_set_ui(numerator, 10UL);
			mpf_add(numerator, numerator, tmp);
			mpf_set_ui(denominator, 72UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 1, 0, coef);
			mpf_set_ui(denominator, 45UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 2, 1, coef);
			mpf_set_ui(numerator, 25UL);
			mpf_sqrt_ui(tmp, 15UL);
			mpf_mul_ui(tmp, tmp, 6UL);
			mpf_sub(numerator, numerator, tmp);
			mpf_set_ui(denominator, 180UL);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 0, 2, coef);
			mpf_set_ui(numerator, 25UL);
			mpf_add(numerator, numerator, tmp);
			mpf_div(coef, numerator, denominator);
			smpfmij(coef_a, 2, 0, coef);

			break;

		default:
			fprintf(stderr, "set_coef: Such a coefficient set not exist.\n");
	}

	mpf_clear(numerator);
	mpf_clear(denominator);
	mpf_clear(coef);
	mpf_clear(tmp);

}
#endif
