/********************************************************************************/
/* taylor.c: Taylor Method for ODEs (Trial Version)                             */
/* Copyright (C) 2011 Tomonori Kouya                                            */
/*                                                                              */
/* This program is free software: you can redistribute it and/or modify it      */
/* under the terms of the GNU Lesser General Public License as published by the */
/* Free Software Foundation, either version 3 of the License or any later       */
/* version.                                                                     */
/*                                                                              */
/* This program is distributed in the hope that it will be useful, but WITHOUT  */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or        */
/* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License */
/* for more details.                                                            */
/*                                                                              */
/* You should have received a copy of the GNU Lesser General Public License     */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                              */
/********************************************************************************/
#include "bncode.h"

/* Taylor Expansion : IEEEE754 Double */

/* Initialize Taylor */
DTaylor init_dtaylor(void)
{
	DTaylor ret;
	long int i;

	if(default_dtaylor_max_degree <= 0)
		return NULL;

	ret = (DTaylor)malloc(sizeof(double_taylor));
	if(ret == NULL)
		return NULL;

	ret->coef = (double *)calloc(sizeof(double), default_dtaylor_max_degree + 1);
	if(ret->coef == NULL)
	{
		free(ret);
		return NULL;
	}

	for(i = 0; i <= default_dtaylor_max_degree; i++)
		*(ret->coef + i) = (double)0.0;

	ret->max_degree = default_dtaylor_max_degree;
	ret->current_degree = 0;

	return ret;
}

/* Initialize Taylor & set constant */
DTaylor init_dtaylor_d(double val)
{
	DTaylor ret;

	ret = init_dtaylor();
	if(ret != NULL)
		*(ret->coef + 0) = val;

	return ret;
}

/* Initialize Taylor & set max_degree */
DTaylor init2_dtaylor(long int max_degree)
{
	DTaylor ret;
	long int i;

	if(max_degree <= 0)
		return NULL;

	ret = (DTaylor)malloc(sizeof(double_taylor));
	if(ret == NULL)
		return NULL;

	ret->coef = (double *)calloc(sizeof(double), max_degree + 1);
	if(ret->coef == NULL)
	{
		free(ret);
		return NULL;
	}

	for(i = 0; i <= max_degree; i++)
		*(ret->coef + i) = (double)0.0;

	ret->max_degree = max_degree;
	ret->current_degree = 0;

	return ret;
}

/* free */
void free_dtaylor(DTaylor op)
{
	free(op->coef);
	free(op);
}

/* Print Taylor */
void print_dtaylor(DTaylor op)
{
	long int i;

	for(i = op->current_degree; i >= 0; i--)
		printf("%5d: %25.17e\n", i, *(op->coef + i));
	printf("\n");
}

/* Copy Taylor */
/* rop := op   */
void copy_dtaylor(DTaylor rop, DTaylor op)
{
	long int i, max_degree;

	max_degree = (rop->max_degree < op->max_degree) ? rop->max_degree : op->max_degree;

	for(i = 0; i <= max_degree; i++)
		*(rop->coef + i) = *(op->coef + i);
	rop->max_degree = max_degree;
	set_current_degree(rop);

}
/* set default max degree */
void set_default_max_degree(long int degree)
{
	default_dtaylor_max_degree = degree;
}

/* set constant */
void set_const_dtaylor(DTaylor rop, double val)
{
	long int i;

	if(rop == NULL)
		return;

	for(i = 1; i <= rop->max_degree; i++)
		*(rop->coef + i) = 0.0;
	*(rop->coef + 0) = val;

	rop->current_degree = 0;

	return;
}

/* factorial */
double dfac(long int n)
{
	double ret;

	ret = (double)n;
	while(--n >= 1)
		ret *= n;

	return ret;
}

/* exp(x) := 1 + x + x^2/2! + ... + x^n/n! */
void set_exp_dtaylor(DTaylor rop)
{
	long int i;
	double tmp_coef, tmp_x, tmp_xpow;

	if(rop == NULL)
		return;

	tmp_coef = 1.0;
	*rop->coef = 1.0;
	for(i = 1; i <= rop->max_degree; i++)
	{
		tmp_coef /= (double)i;
		*(rop->coef + i) = tmp_coef;
	}
	rop->current_degree = rop->max_degree;

	return;
}

/* sin(x) := x - x^3/3! + ... + (-1)^n * x^(2n+1)/(2n+1)! */
void set_sin_dtaylor(DTaylor rop)
{
	long int i;
	double tmp_coef, tmp_sign, tmp_xpow;

	if(rop == NULL)
		return;

	*(rop->coef + 0) = 0.0;
	*(rop->coef + 1) = 1.0;
	tmp_sign = 1.0;
	for(i = 3; i <= rop->max_degree; i += 2)
	{
//		tmp_coef /= (double)i;
//		tmp_coef /= (double)(i-1);
		tmp_sign *= (double)(-1);
		tmp_coef = tmp_sign / dfac(i);
		*(rop->coef + i) = tmp_coef;
		*(rop->coef + i-1) = 0.0;
	}

	for(i = 0; i <= rop->max_degree; i += 2)
		*(rop->coef + i) = 0.0;

	if(rop->max_degree % 2 == 0)
		rop->current_degree = rop->max_degree - 1;
	else
		rop->current_degree = rop->max_degree;

	return;
}

/* cos(x) := 1 - x^2/2! + ... + (-1)^n * x^(2n)/(2n)! */
void set_cos_dtaylor(DTaylor rop)
{
	long int i;
	double tmp_coef, tmp_sign, tmp_xpow;

	if(rop == NULL)
		return;

	*(rop->coef + 0) = 1.0;
	*(rop->coef + 2) = -0.5;
	tmp_sign = -1.0;
	for(i = 4; i <= rop->max_degree; i += 2)
	{
//		tmp_coef /= (double)i;
//		tmp_coef /= (double)(i-1);
		tmp_sign *= (double)(-1);
		tmp_coef = tmp_sign / dfac(i);
		*(rop->coef + i) = tmp_coef;
		*(rop->coef + i-1) = 0.0;
	}
	for(i = 1; i <= rop->max_degree; i += 2)
		*(rop->coef + i) = 0.0;

	if(rop->max_degree % 2 == 0)
		rop->current_degree = rop->max_degree;
	else
		rop->current_degree = rop->max_degree - 1;

	return;
}

/* Set Current Maxmum Degree */
void set_current_degree(DTaylor rop)
{
	long int degree;

	degree = rop->max_degree;

	do {
		if(*(rop->coef + degree) != 0.0)
		{
			rop->current_degree = degree;
			break;
		}
	} while(--degree >= 0);

	if(degree <= 0)
		rop->current_degree = 0;
}

/* Evaluate Taylor */
/* op(x) */
double eval_dtaylor(DTaylor op, double x)
{
	long int deg;
	double ret;

	deg = op->current_degree;
	ret = *(op->coef + deg);
	while(--deg >= 0)
		ret = ret * x + *(op->coef + deg);

	return ret;
}

/* -Taylor */
/* rop(x) = -op(x) */
void neg_dtaylor(DTaylor rop, DTaylor op)
{
	long int deg, i;

	deg = op->current_degree;

	for(i = 0; i <= deg; i++)
		*(rop->coef + i) = -(*(op->coef + i));

	return;
}

/* shift x to x+x0 */
/* rop(x) := op(x+x0) */
void shift_dtaylor(DTaylor rop, DTaylor op, double x0)
{
	long int deg, i, j, k;
	double tmp;

	deg = op->current_degree;

	for(i = deg; i >= 0; i--)
	{
		/* rop[i] := op->coef[i] */
		*(rop->coef + i) = *(op->coef + i);

		/* rop := rop * (x + x0) + op->coef[i] */
		for(j = i; j < deg; j++)
			*(rop->coef + j) += *(rop->coef + j + 1) * x0;

	}
	set_current_degree(rop);
}

/* cos(x) := cos(x0) +(x-x0)*(-sin(x0)) + (x-x0)^2*(-cos(x0))/2! + ... + (-1)^n * x^(2n)/(2n)! */
void set_cos2_dtaylor(DTaylor rop, double x0)
{
	long int i;
	double tmp_coef, tmp_sign, tmp_xpow;

	if(rop == NULL)
		return;

	*rop->coef = cos(x0);
	tmp_sign = 1.0;
	for(i = 1; i <= rop->max_degree; i += 2)
	{
		tmp_sign *= (double)(-1);
		tmp_coef = tmp_sign * sin(x0) / dfac(i);
		*(rop->coef + i) = tmp_coef;
	}
	tmp_sign = 1.0;
	for(i = 2; i <= rop->max_degree; i += 2)
	{
		tmp_sign *= (double)(-1);
		tmp_coef = tmp_sign * cos(x0) / dfac(i);
		*(rop->coef + i) = tmp_coef;
	}

	if(rop->max_degree % 2 == 0)
		rop->current_degree = rop->max_degree;
	else
		rop->current_degree = rop->max_degree - 1;

//	shift_dtaylor(rop, rop, -x0);

	return;
}
/* sin(x) := sin(x0) + (x-x0)*cos(x0) + (x-x0)^2*(-sin(x0))/2! + ... + (-1)^n * x^(2n)/(2n)! */
void set_sin2_dtaylor(DTaylor rop, double x0)
{
	long int i;
	double tmp_coef, tmp_sign, tmp_xpow;

	if(rop == NULL)
		return;

	*rop->coef = sin(x0);
	tmp_sign = -1.0;
	for(i = 1; i <= rop->max_degree; i += 2)
	{
		tmp_sign *= (double)(-1);
		tmp_coef = tmp_sign * cos(x0) / dfac(i);
		*(rop->coef + i) = tmp_coef;
	}
	tmp_sign = 1.0;
	for(i = 2; i <= rop->max_degree; i += 2)
	{
		tmp_sign *= (double)(-1);
		tmp_coef = tmp_sign * sin(x0) / dfac(i);
		*(rop->coef + i) = tmp_coef;
	}

	if(rop->max_degree % 2 == 0)
		rop->current_degree = rop->max_degree;
	else
		rop->current_degree = rop->max_degree - 1;

//	shift_dtaylor(rop, rop, -x0);

	return;
}

/* Add Taylor */
/* rop := op1 + op2 */
void add_dtaylor(DTaylor rop, DTaylor op1, DTaylor op2)
{
	long int i, max_degree, min_degree, flag;

	/* set max_degree */
	if(op1->max_degree < op2->max_degree)
	{
		max_degree = op2->max_degree;
		min_degree = op1->max_degree;
		flag = 2;
	}
	else
	{
		max_degree = op1->max_degree;
		min_degree = op2->max_degree;
		flag = 1;
	}

	if(min_degree >= rop->max_degree)
	{
		for(i = 0; i <= rop->max_degree; i++)
			*(rop->coef + i) = *(op1->coef + i) + *(op2->coef + i);

		set_current_degree(rop);
		return;
	}

	for(i = 0; i <= min_degree; i++)
		*(rop->coef + i) = *(op1->coef + i) + *(op2->coef + i);

	if(max_degree >= rop->max_degree)
	{
		if(flag == 1)
		{
			for(i = min_degree + 1; i <= rop->max_degree; i++)
				*(rop->coef + i) = *(op1->coef + i);
		}
		else if(flag == 2)
		{
			for(i = min_degree + 1; i <= rop->max_degree; i++)
				*(rop->coef + i) = *(op2->coef + i);
		}

		set_current_degree(rop);
		return;
	}

	if(flag == 1)
	{
		for(i = min_degree + 1; i <= max_degree; i++)
			*(rop->coef + i) = *(op1->coef + i);
	}
	else if(flag == 2)
	{
		for(i = min_degree + 1; i <= max_degree; i++)
			*(rop->coef + i) = *(op2->coef + i);
	}

	for(i = max_degree + 1; i <= rop->max_degree; i++)
		*(rop->coef + i) = 0.0;

	set_current_degree(rop);

}

/* Subtract Taylor */
/* rop := op1 - op2 */
void sub_dtaylor(DTaylor rop, DTaylor op1, DTaylor op2)
{
	long int i, max_degree, min_degree, flag;

	/* set max_degree */
	if(op1->max_degree < op2->max_degree)
	{
		max_degree = op2->max_degree;
		min_degree = op1->max_degree;
		flag = 2;
	}
	else
	{
		max_degree = op1->max_degree;
		min_degree = op2->max_degree;
		flag = 1;
	}

	if(min_degree >= rop->max_degree)
	{
		for(i = 0; i <= rop->max_degree; i++)
			*(rop->coef + i) = *(op1->coef + i) - *(op2->coef + i);

		set_current_degree(rop);
		return;
	}

	for(i = 0; i <= min_degree; i++)
		*(rop->coef + i) = *(op1->coef + i) - *(op2->coef + i);

	if(max_degree >= rop->max_degree)
	{
		if(flag == 1)
		{
			for(i = min_degree + 1; i <= rop->max_degree; i++)
				*(rop->coef + i) = *(op1->coef + i);
		}
		else if(flag == 2)
		{
			for(i = min_degree + 1; i <= rop->max_degree; i++)
				*(rop->coef + i) = -*(op2->coef + i);
		}

		set_current_degree(rop);
		return;
	}

	if(flag == 1)
	{
		for(i = min_degree + 1; i <= max_degree; i++)
			*(rop->coef + i) = *(op1->coef + i);
	}
	else if(flag == 2)
	{
		for(i = min_degree + 1; i <= max_degree; i++)
			*(rop->coef + i) = -*(op2->coef + i);
	}

	for(i = max_degree + 1; i <= rop->max_degree; i++)
		*(rop->coef + i) = 0.0;

	set_current_degree(rop);

}

/* Scalar Mulutiply Taylor */
/* rop := const * op */
void smul_dtaylor(DTaylor rop, double sc, DTaylor op)
{
	long int i, min_degree;

	if(op->max_degree <= rop->max_degree)
		min_degree = op->max_degree;
	else
		min_degree = rop->max_degree;

	for(i = 0; i <= min_degree; i++)
		*(rop->coef + i) = sc * *(op->coef + i);

	for(i = min_degree + 1; i <= rop->max_degree; i++)
		*(rop->coef + i) = 0.0;

	set_current_degree(rop);
}

/* Scalar Divide Taylor */
/* rop := const / op */
void sdiv_dtaylor(DTaylor rop, DTaylor op, double sc)
{
	long int i, min_degree;

	if(op->max_degree <= rop->max_degree)
		min_degree = op->max_degree;
	else
		min_degree = rop->max_degree;

	for(i = 0; i <= min_degree; i++)
		*(rop->coef + i) = *(op->coef + i) / sc;

	for(i = min_degree + 1; i <= rop->max_degree; i++)
		*(rop->coef + i) = 0.0;

	set_current_degree(rop);
}

/* Mulutiply Taylor */
/* rop := op1 * op2 */
void mul_dtaylor(DTaylor rop, DTaylor op1, DTaylor op2)
{
	long int i, j, min_degree, op1_deg, op2_deg;
	DTaylor tmp_taylor;
	double tmp;

	tmp_taylor = init2_dtaylor(rop->max_degree);

	min_degree = op1->current_degree + op2->current_degree;
	if(min_degree > rop->max_degree)
		min_degree = rop->max_degree;

	for(i = 0; i <= min_degree; i++)
	{
		tmp = 0.0;
		op1_deg = (i <= op1->current_degree) ? i : op1->current_degree;
		for(j = op1_deg; j >= 0; j--)
		{
			if((i - j) > op2->current_degree)
				break;
			tmp += *(op1->coef + j) * *(op2->coef + i - j);
		}

		*(tmp_taylor->coef + i) = tmp;
	}
	copy_dtaylor(rop, tmp_taylor);

	set_current_degree(rop);
	free_dtaylor(tmp_taylor);
}

/* Integral Taylor */
/* rop := Int[x0, x](op) */
void integral_dtaylor(DTaylor rop, DTaylor op, double x0)
{
	long int i, i1, min_degree;
	double tmp_coef;

	if(op->max_degree <= rop->max_degree)
		min_degree = op->max_degree;
	else
		min_degree = rop->max_degree;

	for(i = min_degree-1; i >= 0; i--)
	{
		i1 = i + 1;
		tmp_coef = *(op->coef + i) / (double)i1;
		*(rop->coef + i1) = tmp_coef;
	}
	*rop->coef = 0.0;

	for(i = min_degree + 1; i <= rop->max_degree; i++)
		*(rop->coef + i) = 0.0;

	set_current_degree(rop);

	/* int(p(t), [x0, x]) = [P(t)]^x_x0 = P(x) - P(x0) */
	*(rop->coef + 0) = -eval_dtaylor(rop, x0);

}

/* Integral2 Taylor */
/* rop := Int[x0, x0+x](op) */
void integral2_dtaylor(DTaylor rop, DTaylor op, double x0)
{
	long int i, i1, min_degree;
	double tmp_coef, tmp;

	if(op->max_degree <= rop->max_degree)
		min_degree = op->max_degree;
	else
		min_degree = rop->max_degree;

//	min_degree = op->current_degree;
	for(i = min_degree-1; i >= 0; i--)
	{
		i1 = i + 1;
		tmp_coef = *(op->coef + i) / (double)i1;
		*(rop->coef + i1) = tmp_coef;
	}
	*rop->coef = 0.0;
//	rop->current_degree = op->current_degree + 1;

	for(i = min_degree ; i <= rop->max_degree; i++)
		*(rop->coef + i) = 0.0;

	set_current_degree(rop);

	/* int(p(t), [x0, x0+x]) = [P(t)]^x0_x0+x = P(x+x0) - P(x0) */
	tmp = eval_dtaylor(rop, x0);
	shift_dtaylor(rop, rop, x0);
	*rop->coef -= tmp;

}

/* Differential Taylor */
/* rop := Diff(op) */
void diff_dtaylor(DTaylor rop, DTaylor op)
{
	long int i, i1, min_degree;
	double tmp_coef;

	if(op->max_degree <= rop->max_degree)
		min_degree = op->max_degree;
	else
		min_degree = rop->max_degree;

	for(i = 1; i <= min_degree; i++)
	{
		i1 = i - 1;
		tmp_coef = *(op->coef + i) * (double)i;
		*(rop->coef + i1) = tmp_coef;
	}

	for(i = min_degree + 1; i <= rop->max_degree; i++)
		*(rop->coef + i) = 0.0;

	set_current_degree(rop);
}

#ifdef USE_GMP

/* Initialize Taylor */
MPFTaylor init_mpftaylor(void)
{
	MPFTaylor ret;
	long int i;

	if(default_mpftaylor_max_degree <= 0)
		return NULL;

	ret = (MPFTaylor)malloc(sizeof(mpf_taylor));
	if(ret == NULL)
		return NULL;

	ret->coef = (mpf_t *)calloc(sizeof(mpf_t), default_mpftaylor_max_degree + 1);
	if(ret->coef == NULL)
	{
		free(ret);
		return NULL;
	}


	ret->prec = bnc_default_prec;
	ret->max_degree = default_mpftaylor_max_degree;
	ret->current_degree = 0;

	for(i = 0; i <= ret->max_degree; i++)
		mpf_init_set_ui(*(ret->coef + i), 0UL);

	return ret;
}

/* Initialize Taylor & set constant */
MPFTaylor init_mpftaylor_d(mpf_t val)
{
	MPFTaylor ret;

	ret = init_mpftaylor();
	if(ret != NULL)
		mpf_set(*(ret->coef + 0), val);

	return ret;
}

/* Initialize Taylor & set constant */
MPFTaylor init_mpftaylor_str(const char *val)
{
	MPFTaylor ret;

	ret = init_mpftaylor();
	if(ret != NULL)
		mpf_set_str(*(ret->coef + 0), val, 10);

	return ret;
}

/* Initialize Taylor & set max_degree */
MPFTaylor init2_mpftaylor(long int max_degree)
{
	MPFTaylor ret;
	long int i;

	if(max_degree <= 0)
		return NULL;

	ret = (MPFTaylor)malloc(sizeof(mpf_taylor));
	if(ret == NULL)
		return NULL;

	ret->coef = (mpf_t *)calloc(sizeof(mpf_t), max_degree + 1);
	if(ret->coef == NULL)
	{
		free(ret);
		return NULL;
	}

	for(i = 0; i <= max_degree; i++)
		mpf_init_set_ui(*(ret->coef + i), 0UL);

	ret->max_degree = max_degree;
	ret->prec = bnc_default_prec;
	ret->current_degree = 0;

	return ret;
}

/* free */
void free_mpftaylor(MPFTaylor op)
{
	long int i;

	for(i = 0; i <= op->max_degree; i++)
		mpf_clear(*(op->coef + i));
	free(op->coef);
	free(op);
}

/* Print Taylor */
void print_mpftaylor(MPFTaylor op)
{
	long int i, deg;

	for(i = op->current_degree; i >= 0; i--)
	{
		printf("%5d: ", i);
		mpf_out_str(stdout, 10, 0, *(op->coef + i) );
		printf("\n");
	}
	printf("\n");
}

/* Copy Taylor */
/* rop := op   */
void copy_mpftaylor(MPFTaylor rop, MPFTaylor op)
{
	long int i, max_degree;

	max_degree = (rop->max_degree < op->max_degree) ? rop->max_degree : op->max_degree;

	for(i = 0; i <= max_degree; i++)
		mpf_set(*(rop->coef + i), *(op->coef + i));
	rop->max_degree = max_degree;
	set_current_degree_mpftaylor(rop);

}
/* set default max degree */
void set_default_max_degree_mpftaylor(long int degree)
{
	default_mpftaylor_max_degree = degree;
}

/* set constant */
void set_const_mpftaylor(MPFTaylor rop, mpf_t val)
{
	long int i;

	if(rop == NULL)
		return;

	for(i = 1; i <= rop->max_degree; i++)
		mpf_set_ui(*(rop->coef + i), 0UL);
	mpf_set(*(rop->coef + 0), val);

	rop->current_degree = 0;

	return;
}
/* set constant */
void set_const_mpftaylor_str(MPFTaylor rop, const char *val)
{
	long int i;

	if(rop == NULL)
		return;

	for(i = 1; i <= rop->max_degree; i++)
		mpf_set_ui(*(rop->coef + i), 0UL);
	mpf_set_str(*(rop->coef + 0), val, 10);

	rop->current_degree = 0;

	return;
}

/* factorial */
void mpf_fac(mpf_t ret, long int n)
{

	mpf_set_ui(ret, (unsigned long)n);
	while(--n >= 1)
		mpf_mul_ui(ret, ret, n);
}

/* exp(x) := 1 + x + x^2/2! + ... + x^n/n! */
void set_exp_mpftaylor(MPFTaylor rop)
{
	long int i;
	mpf_t tmp_coef, tmp_x, tmp_xpow;

	if(rop == NULL)
		return;

	mpf_set_ui(*rop->coef, 1UL);
	for(i = 1; i <= rop->max_degree; i++)
	{
		mpf_fac(tmp_coef, i);
		mpf_ui_div(tmp_coef, 1UL, tmp_coef);
		mpf_set(*(rop->coef + i), tmp_coef);
	}
	rop->current_degree = rop->max_degree;

	return;
}

/* sin(x) := x - x^3/3! + ... + (-1)^n * x^(2n+1)/(2n+1)! */
void set_sin_mpftaylor(MPFTaylor rop)
{
	long int i, tmp_sign;
	mpf_t tmp_coef, tmp_coef2;

	if(rop == NULL)
		return;

	mpf_init2(tmp_coef, rop->prec);
	mpf_init2(tmp_coef2, rop->prec);

	mpf_set_ui(*(rop->coef + 0), 0UL);
	mpf_set_ui(*(rop->coef + 1), 1UL);
	tmp_sign = 1;
	for(i = 3; i <= rop->max_degree; i += 2)
	{
//		tmp_coef /= (mpf_t)i;
//		tmp_coef /= (mpf_t)(i-1);
		mpf_fac(tmp_coef2, (unsigned long)i);
		mpf_ui_div(tmp_coef, 1UL, tmp_coef2);
		tmp_sign *= -1;
		if(tmp_sign == -1)
			mpf_neg(tmp_coef, tmp_coef);
		mpf_set(*(rop->coef + i), tmp_coef);
		mpf_set_ui(*(rop->coef + i-1), 0UL);
	}

	if(rop->max_degree % 2 == 0)
		rop->current_degree = rop->max_degree - 1;
	else
		rop->current_degree = rop->max_degree;

	mpf_clear(tmp_coef);
	mpf_clear(tmp_coef2);

	return;
}

/* cos(x) := 1 - x^2/2! + ... + (-1)^n * x^(2n)/(2n)! */
void set_cos_mpftaylor(MPFTaylor rop)
{
	long int i, tmp_sign;
	mpf_t tmp_coef, tmp_coef2;

	if(rop == NULL)
		return;

	mpf_init2(tmp_coef, rop->prec);
	mpf_init2(tmp_coef2, rop->prec);

	mpf_set_ui(*(rop->coef + 0), 1UL);
	mpf_set_str(*(rop->coef + 2), "-0.5", 10);
	tmp_sign = -1;
	for(i = 4; i <= rop->max_degree; i += 2)
	{
//		tmp_coef /= (mpf_t)i;
//		tmp_coef /= (mpf_t)(i-1);
		tmp_sign *= -1;
		mpf_fac(tmp_coef2, i);
		mpf_ui_div(tmp_coef, 1UL, tmp_coef2);
		if(tmp_sign == -1)
			mpf_neg(tmp_coef, tmp_coef);
		mpf_set(*(rop->coef + i), tmp_coef);
		mpf_set_ui(*(rop->coef + i-1), 0UL);
	}

	if(rop->max_degree % 2 == 0)
		rop->current_degree = rop->max_degree;
	else
		rop->current_degree = rop->max_degree - 1;

	mpf_clear(tmp_coef);
	mpf_clear(tmp_coef2);

	return;
}

/* Set Current Maxmum Degree */
void set_current_degree_mpftaylor(MPFTaylor rop)
{
	long int degree;

	degree = rop->max_degree;

	do {
		if(mpf_cmp_ui(*(rop->coef + degree), 0UL) != 0)
		{
			rop->current_degree = degree;
			break;
		}
	} while(--degree >= 0);

	if(degree <= 0)
		rop->current_degree = 0;
}

/* Evaluate Taylor */
/* op(x) */
void eval_mpftaylor(mpf_t ret, MPFTaylor op, mpf_t x)
{
	long int deg;

	deg = op->current_degree;
	mpf_set(ret, *(op->coef + deg));
	while(--deg >= 0)
	{
		mpf_mul(ret, ret, x);
		mpf_add(ret, ret, *(op->coef + deg));
	}

}

/* -Taylor */
/* rop(x) = -op(x) */
void neg_mpftaylor(MPFTaylor rop, MPFTaylor op)
{
	long int deg, i;

	deg = op->current_degree;

	for(i = 0; i <= deg; i++)
		mpf_neg(*(rop->coef + i), *(op->coef + i));

	return;
}

/* shift x to x+x0 */
/* rop(x) := op(x+x0) */
void shift_mpftaylor(MPFTaylor rop, MPFTaylor op, mpf_t x0)
{
	long int deg, i, j, k;
	mpf_t tmp;

	mpf_init2(tmp, rop->prec);
	deg = op->current_degree;

	for(i = deg; i >= 0; i--)
	{
		/* rop[i] := op->coef[i] */
		mpf_set(*(rop->coef + i), *(op->coef + i));

		/* rop := rop * (x + x0) + op->coef[i] */
		for(j = i; j < deg; j++)
		{
			mpf_mul(tmp, *(rop->coef + j + 1), x0);
			mpf_add(*(rop->coef + j), *(rop->coef + j), tmp);
		}

	}
	set_current_degree_mpftaylor(rop);
	mpf_clear(tmp);
}

/* factorial */
/* void mpf_fac(mpf_t ret, long int n)
{
	mpf_set_ui(ret, (unsigned long)n);
	while(--n >= 1)
		mpf_mul_ui(ret, ret, (unsigned long)n);
}
*/

/* cos(x) := cos(x0) +(x-x0)*(-sin(x0)) + (x-x0)^2*(-cos(x0))/2! + ... + (-1)^n * x^(2n)/(2n)! */
void set_cos2_mpftaylor(MPFTaylor rop, mpf_t x0)
{
	long int i, tmp_sign;
	mpf_t tmp_coef, tmp_coef2;

	if(rop == NULL)
		return;

	mpf_init2(tmp_coef, rop->prec);
	mpf_init2(tmp_coef2, rop->prec);

	mpf_cos(*rop->coef, x0);
	tmp_sign = 1;
	for(i = 1; i <= rop->max_degree; i += 2)
	{
		tmp_sign *= -1;
		mpf_sin(tmp_coef, x0);
		mpf_fac(tmp_coef2, i);
		mpf_div(tmp_coef, tmp_coef, tmp_coef2);
		if(tmp_sign < 0)
			mpf_neg(tmp_coef, tmp_coef);
		mpf_set(*(rop->coef + i), tmp_coef);
	}
	tmp_sign = 1;
	for(i = 2; i <= rop->max_degree; i += 2)
	{
		tmp_sign *= -1;
		mpf_cos(tmp_coef, x0);
		mpf_fac(tmp_coef2, i);
		mpf_div(tmp_coef, tmp_coef, tmp_coef2);
		if(tmp_sign < 0)
			mpf_neg(tmp_coef, tmp_coef);
		mpf_set(*(rop->coef + i), tmp_coef);
	}

	if(rop->max_degree % 2 == 0)
		rop->current_degree = rop->max_degree;
	else
		rop->current_degree = rop->max_degree - 1;

//	shift_dtaylor(rop, rop, -x0);

	mpf_clear(tmp_coef);
	mpf_clear(tmp_coef2);

	return;
}

/* sin(x) := sin(x0) + (x-x0)*cos(x0) + (x-x0)^2*(-sin(x0))/2! + ... + (-1)^n * x^(2n)/(2n)! */
void set_sin2_mpftaylor(MPFTaylor rop, mpf_t x0)
{
	long int i, tmp_sign;
	mpf_t tmp_coef, tmp_coef2;

	if(rop == NULL)
		return;

	mpf_init2(tmp_coef, rop->prec);
	mpf_init2(tmp_coef2, rop->prec);

	mpf_sin(*rop->coef, x0);
	tmp_sign = -1;
	for(i = 1; i <= rop->max_degree; i += 2)
	{
		tmp_sign *= -1;
		mpf_cos(tmp_coef, x0);
		mpf_fac(tmp_coef2, i);
		mpf_div(tmp_coef, tmp_coef, tmp_coef2);
		if(tmp_sign < 0)
			mpf_neg(tmp_coef, tmp_coef);
		mpf_set(*(rop->coef + i), tmp_coef);
	}
	tmp_sign = 1;
	for(i = 2; i <= rop->max_degree; i += 2)
	{
		tmp_sign *= -1;
		mpf_sin(tmp_coef, x0);
		mpf_fac(tmp_coef2, i);
		mpf_div(tmp_coef, tmp_coef, tmp_coef2);
		if(tmp_sign < 0)
			mpf_neg(tmp_coef, tmp_coef);
		mpf_set(*(rop->coef + i), tmp_coef);
	}

	if(rop->max_degree % 2 == 0)
		rop->current_degree = rop->max_degree;
	else
		rop->current_degree = rop->max_degree - 1;

//	shift_dtaylor(rop, rop, -x0);

	mpf_clear(tmp_coef);
	mpf_clear(tmp_coef2);

	return;
}

/* Add Taylor */
/* rop := op1 + op2 */
void add_mpftaylor(MPFTaylor rop, MPFTaylor op1, MPFTaylor op2)
{
	long int i, max_degree, min_degree, flag;

	/* set max_degree */
	if(op1->max_degree < op2->max_degree)
	{
		max_degree = op2->max_degree;
		min_degree = op1->max_degree;
		flag = 2;
	}
	else
	{
		max_degree = op1->max_degree;
		min_degree = op2->max_degree;
		flag = 1;
	}

	if(min_degree >= rop->max_degree)
	{
		for(i = 0; i <= rop->max_degree; i++)
			mpf_add(*(rop->coef + i), *(op1->coef + i), *(op2->coef + i));

		set_current_degree_mpftaylor(rop);
		return;
	}

	for(i = 0; i <= min_degree; i++)
		mpf_add(*(rop->coef + i), *(op1->coef + i), *(op2->coef + i));

	if(max_degree >= rop->max_degree)
	{
		if(flag == 1)
		{
			for(i = min_degree + 1; i <= rop->max_degree; i++)
				mpf_set(*(rop->coef + i), *(op1->coef + i));
		}
		else if(flag == 2)
		{
			for(i = min_degree + 1; i <= rop->max_degree; i++)
				mpf_set(*(rop->coef + i), *(op2->coef + i));
		}

		set_current_degree_mpftaylor(rop);
		return;
	}

	if(flag == 1)
	{
		for(i = min_degree + 1; i <= max_degree; i++)
			mpf_set(*(rop->coef + i), *(op1->coef + i));
	}
	else if(flag == 2)
	{
		for(i = min_degree + 1; i <= max_degree; i++)
			mpf_set(*(rop->coef + i), *(op2->coef + i));
	}

	for(i = max_degree + 1; i <= rop->max_degree; i++)
		mpf_set_ui(*(rop->coef + i), 0UL);

	set_current_degree_mpftaylor(rop);

}

/* Subtract Taylor */
/* rop := op1 - op2 */
void sub_mpftaylor(MPFTaylor rop, MPFTaylor op1, MPFTaylor op2)
{
	long int i, max_degree, min_degree, flag;

	/* set max_degree */
	if(op1->max_degree < op2->max_degree)
	{
		max_degree = op2->max_degree;
		min_degree = op1->max_degree;
		flag = 2;
	}
	else
	{
		max_degree = op1->max_degree;
		min_degree = op2->max_degree;
		flag = 1;
	}

	if(min_degree >= rop->max_degree)
	{
		for(i = 0; i <= rop->max_degree; i++)
			mpf_sub(*(rop->coef + i), *(op1->coef + i), *(op2->coef + i));

		set_current_degree_mpftaylor(rop);
		return;
	}

	for(i = 0; i <= min_degree; i++)
		mpf_sub(*(rop->coef + i), *(op1->coef + i), *(op2->coef + i));

	if(max_degree >= rop->max_degree)
	{
		if(flag == 1)
		{
			for(i = min_degree + 1; i <= rop->max_degree; i++)
				mpf_set(*(rop->coef + i), *(op1->coef + i));
		}
		else if(flag == 2)
		{
			for(i = min_degree + 1; i <= rop->max_degree; i++)
				mpf_neg(*(rop->coef + i), *(op2->coef + i));
		}

		set_current_degree_mpftaylor(rop);
		return;
	}

	if(flag == 1)
	{
		for(i = min_degree + 1; i <= max_degree; i++)
			mpf_set(*(rop->coef + i), *(op1->coef + i));
	}
	else if(flag == 2)
	{
		for(i = min_degree + 1; i <= max_degree; i++)
			mpf_neg(*(rop->coef + i), *(op2->coef + i));
	}

	for(i = max_degree + 1; i <= rop->max_degree; i++)
		mpf_set_ui(*(rop->coef + i), 0UL);

	set_current_degree_mpftaylor(rop);

}

/* Scalar Mulutiply Taylor */
/* rop := const * op */
void smul_mpftaylor(MPFTaylor rop, mpf_t sc, MPFTaylor op)
{
	long int i, min_degree;

	if(op->max_degree <= rop->max_degree)
		min_degree = op->max_degree;
	else
		min_degree = rop->max_degree;

	for(i = 0; i <= min_degree; i++)
		mpf_mul(*(rop->coef + i), sc, *(op->coef + i));

	for(i = min_degree + 1; i <= rop->max_degree; i++)
		mpf_set_ui(*(rop->coef + i), 0UL);

	set_current_degree_mpftaylor(rop);
}

/* Scalar Divide Taylor */
/* rop := const / op */
void sdiv_mpftaylor(MPFTaylor rop, MPFTaylor op, mpf_t sc)
{
	long int i, min_degree;

	if(op->max_degree <= rop->max_degree)
		min_degree = op->max_degree;
	else
		min_degree = rop->max_degree;

	for(i = 0; i <= min_degree; i++)
		mpf_div(*(rop->coef + i), *(op->coef + i), sc);

	for(i = min_degree + 1; i <= rop->max_degree; i++)
		mpf_set_ui(*(rop->coef + i), 0UL);

	set_current_degree_mpftaylor(rop);
}

/* Mulutiply Taylor */
/* rop := op1 * op2 */
void mul_mpftaylor(MPFTaylor rop, MPFTaylor op1, MPFTaylor op2)
{
	long int i, j, min_degree, op1_deg, op2_deg;
	mpf_t tmp, tmp2;
	MPFTaylor tmp_taylor;

	mpf_init2(tmp, rop->prec);
	mpf_init2(tmp2, rop->prec);
	tmp_taylor = init2_mpftaylor(rop->max_degree);

	min_degree = op1->current_degree + op2->current_degree;
	if(min_degree > rop->max_degree)
		min_degree = rop->max_degree;

	for(i = 0; i <= min_degree; i++)
	{
		mpf_set_ui(tmp, 0UL);
		op1_deg = (i <= op1->current_degree) ? i : op1->current_degree;
		for(j = op1_deg; j >= 0; j--)
		{
			if((i - j) > op2->current_degree)
				break;
			mpf_mul(tmp2, *(op1->coef + j), *(op2->coef + i - j));
			mpf_add(tmp, tmp, tmp2);
		}

		mpf_set(*(tmp_taylor->coef + i), tmp);
	}
	copy_mpftaylor(rop, tmp_taylor);

	set_current_degree_mpftaylor(rop);
	mpf_clear(tmp);
	mpf_clear(tmp2);
	free_mpftaylor(tmp_taylor);
}

/* Integral Taylor */
/* rop := Int[x0, x](op) */
void integral_mpftaylor(MPFTaylor rop, MPFTaylor op, mpf_t x0)
{
	long int i, i1, min_degree;
	mpf_t tmp_coef;

	mpf_init2(tmp_coef, rop->prec);

	if(op->max_degree <= rop->max_degree)
		min_degree = op->max_degree;
	else
		min_degree = rop->max_degree;

	for(i = min_degree-1; i >= 0; i--)
	{
		i1 = i + 1;
		mpf_div_ui(tmp_coef, *(op->coef + i), i1);
		mpf_set(*(rop->coef + i1), tmp_coef);
	}
	mpf_set_ui(*rop->coef, 0UL);

	for(i = min_degree + 1; i <= rop->max_degree; i++)
		mpf_set_ui(*(rop->coef + i), 0UL);

	set_current_degree_mpftaylor(rop);

	/* int(p(t), [x0, x]) = [P(t)]^x_x0 = P(x) - P(x0) */
	eval_mpftaylor(tmp_coef, rop, x0);
	mpf_neg(*(rop->coef + 0), tmp_coef);

}
#endif
