/********************************************************************************/
/*                                                                              */
/* bncsparse.h : Header file for sparse_double.c and sparse_mpf.c               */
/* Copyright (c) 2006-2011 Tomonori Kouya, All rights reserved.                 */
/*                                                                              */
/* 2007-11-13 Version 0.1: Bug fix of get_secv,                                 */
/*     append get_dvector_i, set_dvector_i, add_dvector                         */
/*     append nzero_row_dim[] and set_nzero_row_dim                             */
/*                                                                              */
/* 2011-06-28 Version 0.2: Reconstruction of sparse.h to be merged into BNCpack */
/* 2011-08-29 Version 0.3: Bug fix in sparse_double.c & sparse_mpf.c            */
/* 2011-03-20 Version 0.31:Add "zero_element" to *RSMatrix structure            */
/*                                                                              */
/* This program is free software: you can redistribute it and/or modify it      */
/* under the terms of the GNU Lesser General Public License as published by the */
/* Free Software Foundation, either version 3 of the License or any later       */
/* version.                                                                     */
/*                                                                              */
/* This program is distributed in the hope that it will be useful, but WITHOUT  */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or        */
/* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License */
/* for more details.                                                            */
/*                                                                              */
/* You should have received a copy of the GNU Lesser General Public License     */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                              */
/********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "bnc.h"

#ifndef __BNC_SPARSE_H__
#define __BNC_SPARSE_H__

#define ERROR   (-1)
#define SUCCESS (0)

#define EMPTY (-1)

#define LINE_BUF_LEN 128

/********************/
/* double precision */
/********************/

/* Sparse matrix struct */
/* Example:             */
/*      0 1 2 3 4       */
/* A = [a 0 b c 0]0     */
/*     [0 d 0 0 0]1     */
/*     [0 e f 0 0]2     */
/*     [0 0 0 g 0]3     */
/*     [0 0 h 0 i]4     */
/*                      */
/* <--> element = [a b c d e f g h i] */
/*      row_dim = 5, col_dim = 5      */
/*      nzero_index[0] = [0 2 3]    */
/*      nzero_index[1] = [4]        */
/*      nzero_index[2] = [1 2]      */
/*      nzero_index[3] = [3]        */
/*      nzero_index[4] = [2 4]      */
/*      nzero_col_dim[0] = 3 */
/*      nzero_col_dim[1] = 1 */
/*      nzero_col_dim[2] = 2 */
/*      nzero_col_dim[3] = 1 */
/*      nzero_col_dim[4] = 2 */
/*      nzero_total_num = 9 */
/*      nzero_row_dim[0] = 1 */
/*      nzero_row_dim[1] = 2 */
/*      nzero_row_dim[2] = 3 */
/*      nzero_row_dim[3] = 2 */
/*      nzero_row_dim[4] = 1 */

typedef struct {
	double *element;			// Elements of matrix
	long int row_dim, col_dim;	// Dimensions of Row and Column
	long int **nzero_index;		// Indeces of Non-zero elements
	long int *nzero_col_dim;	// Numbers of non-zero elements in i-th row
	long int *nzero_row_dim;	// Numbers of non-zero elements in i-th column
	long int nzero_total_num;	// Total number of non-zero elements
	double zero_element;        // Zero value (Append in 2012-03-20)
} drsmatrix;

/* Definition of "DRSMatrix" type */
typedef drsmatrix *DRSMatrix;

/* initialize DRSMatrix */
DRSMatrix init_drsmatrix(long int, long int *, long int);

/* Clear DRSMatrix */
void free_drsmatrix(DRSMatrix);

/* set nzero_row_dim automatically */
void set_nzero_row_dim(DRSMatrix);

/* Print DRSMatrix */
void print_drsmatrix(DRSMatrix);

/* initialize and substitute DRSMatrix from DMatrix */
DRSMatrix init_set_drsmatrix_dmatrix(DMatrix);

/* Get variables to initialize DRSMatrix */
int get_vars_drsmatrix_fname(long int *, long int **, long int *, const char *);

/* Read URI linking data */
int fread_urilinkdat_fname(DRSMatrix, const char *);

/* Multiply DRSMatrix * DVector */
int mul_drsmatrix_dvec(DVector, DRSMatrix, DVector);

/* Multiply DRSMatrix^T * DVector */
int mul_drsmatrixt_dvec(DVector, DRSMatrix, DVector);

/* Power Method for Randomly Sparse Matrices */
/* 	double *evec: the eigenvector for max eigenvalue */
/* 	double *drsmat: Randomly sparse matrix */
/* 	double reps, aeps: Relative and Absolute tolerance */
/* 	long int max_times: Maximum iterative times of Power method */
double dpower_rsmatrix(DVector, DRSMatrix, double, double, long int);

/* Scalar multiply of DVector */
int smul_dvector(DVector, double, DVector);

/* Select index of absolute maximum element and its value in DVector */
long int absmax_index_dvector(double *, DVector);

/* MPF */
#ifdef USE_GMP

/* Sparse matrix struct */
/* Example:             */
/*      0 1 2 3 4       */
/* A = [a 0 b c 0]0     */
/*     [0 d 0 0 0]1     */
/*     [0 e f 0 0]2     */
/*     [0 0 0 g 0]3     */
/*     [0 0 h 0 i]4     */
/*                      */
/* <--> element = [a b c d e f g h i] */
/*      row_dim = 5, col_dim = 5      */
/*      nzero_index[0] = [0 2 3]    */
/*      nzero_index[1] = [4]        */
/*      nzero_index[2] = [1 2]      */
/*      nzero_index[3] = [3]        */
/*      nzero_index[4] = [2 4]      */
/*      nzero_col_dim[0] = 3 */
/*      nzero_col_dim[1] = 1 */
/*      nzero_col_dim[2] = 2 */
/*      nzero_col_dim[3] = 1 */
/*      nzero_col_dim[4] = 2 */
/*      nzero_total_num =  9 */
/*      nzero_row_dim[0] = 1 */
/*      nzero_row_dim[1] = 2 */
/*      nzero_row_dim[2] = 3 */
/*      nzero_row_dim[3] = 2 */
/*      nzero_row_dim[4] = 1 */

typedef struct {
	unsigned long prec;
	mpf_t *element;				// Elements of matrix
	long int row_dim, col_dim;	// Dimensions of Row and Column
	long int **nzero_index;		// Indeces of Non-zero elements
	long int *nzero_col_dim;	// Numbers of non-zero elements in i-th row
	long int *nzero_row_dim;	// Numbers of non-zero elements in i-th column
	long int nzero_total_num;	// Total number of non-zero elements
	mpf_t zero_element;         // Zero value (Append in 2012-03-20)
} mpfrsmatrix;

/* Definition of "MPFRSMatrix" type */
typedef mpfrsmatrix *MPFRSMatrix;

/* initialize MPFRSMatrix */
MPFRSMatrix init_mpfrsmatrix(long int, long int *, long int);

/* initialize MPFRSMatrix with specific precision */
MPFRSMatrix init2_mpfrsmatrix(long int, long int *, long int, unsigned long prec);

/* Clear MPFRSMatrix */
void free_mpfrsmatrix(MPFRSMatrix);

/* set nzero_row_dim automatically */
void set_nzero_row_dim_mpf(MPFRSMatrix);

/* Print MPFRSMatrix */
void print_mpfrsmatrix(MPFRSMatrix);

/* initialize and substitute MPFRSMatrix from MPFMatrix */
MPFRSMatrix init_set_mpfrsmatrix_mpfmatrix(MPFMatrix);

/* Get variables to initialize MPFRSMatrix */
int get_vars_mpfrsmatrix_fname(long int *, long int **, long int *, const char *);

/* Read URI linking data */
int fread_urilinkdat_fname_mpf(MPFRSMatrix, const char *);

/* Select index of absolute maximum element and its value in MPFVector */
long int absmax_index_mpfvector(mpf_t, MPFVector);

/* Multiply MPFRSMatrix * MPFVector */
int mul_mpfrsmatrix_mpfvec(MPFVector, MPFRSMatrix, MPFVector);

/* Multiply MPFRSMatrix^T * MPFVector */
int mul_mpfrsmatrixt_mpfvec(MPFVector, MPFRSMatrix, MPFVector);

/* Power Method for Randomly Sparse Matrices */
/* 	double *evec: the eigenvector for max eigenvalue */
/* 	double *drsmat: Randomly sparse matrix */
/* 	double reps, aeps: Relative and Absolute tolerance */
/* 	long int max_times: Maximum iterative times of Power method */
void mpfpower_rsmatrix(mpf_t, MPFVector, MPFRSMatrix, mpf_t, mpf_t, long int);

/* Scalar multiply of MPFVector */
int smul_mpfvector(MPFVector, mpf_t, MPFVector);

/* Select index of absolute maximum element and its value in MPFVector */
long int absmax_index_mpfvector(mpf_t, MPFVector);

#endif

/***********************************************/
/* Conjugate-Gradient Method (Sparse Version) */
/**********************************************/
long int DCG_sp(DVector answer, DRSMatrix a, DVector b, double reps, double aeps, long int maxtimes);
#ifdef USE_GMP
long int MPFCG_sp(MPFVector answer, MPFRSMatrix a, MPFVector b, mpf_t reps, mpf_t aeps, long int maxtimes);
#endif

/********************************************/
/* Krylov Subspace Methods (Sparse Version) */
/********************************************/
long int DBiCG_sp(DVector answer, DRSMatrix a, DVector b, double reps, double aeps, long int maxtimes);
long int DCGS_sp(DVector answer, DRSMatrix a, DVector b, double reps, double aeps, long int maxtimes);
long int DBiCGSTAB_sp(DVector answer, DRSMatrix a, DVector b, double reps, double aeps, long int maxtimes);
long int DGPBiCG_sp(DVector answer, DRSMatrix a, DVector b, double reps, double aeps, long int maxtimes);

#ifdef USE_GMP
long int MPFBiCG_sp(MPFVector answer, MPFRSMatrix a, MPFVector b, mpf_t reps, mpf_t aeps, long int maxtimes);
long int MPFCGS_sp(MPFVector answer, MPFRSMatrix a, MPFVector b, mpf_t reps, mpf_t aeps, long int maxtimes);
long int MPFBiCGSTAB_sp(MPFVector answer, MPFRSMatrix a, MPFVector b, mpf_t reps, mpf_t aeps, long int maxtimes);
long int MPFGPBiCG_sp(MPFVector answer, MPFRSMatrix a, MPFVector b, mpf_t reps, mpf_t aeps, long int maxtimes);
#endif
/* End of __BNC_SPARSE_H__ */
#endif
