/********************************************************************************/
/*                                                                              */
/* sparse_double.c : Sparse Matrix and Vector Library (Double Precision)        */
/* Copyright (c) 2006-2011 Tomonori Kouya, All rights reserved.                 */
/*                                                                              */
/* Version 0.1 2011-08-29 : Bug fix in mul_mpfrsmatrix_mpfvec                   */
/*                                                                              */
/* This program is free software: you can redistribute it and/or modify it      */
/* under the terms of the GNU Lesser General Public License as published by the */
/* Free Software Foundation, either version 3 of the License or any later       */
/* version.                                                                     */
/*                                                                              */
/* This program is distributed in the hope that it will be useful, but WITHOUT  */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or        */
/* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License */
/* for more details.                                                            */
/*                                                                              */
/* You should have received a copy of the GNU Lesser General Public License     */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                              */
/********************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef WIN32
#include <windows.h>
#else
#include <sys/times.h>
#include <time.h>
#endif

#include "bncsparse.h"

/* Sparse matrix struct */
/* Example:             */
/*      0 1 2 3 4       */
/* A = [a 0 b c 0]0     */
/*     [0 d 0 0 0]1     */
/*     [0 e f 0 0]2     */
/*     [0 0 0 g 0]3     */
/*     [0 0 h 0 i]4     */
/*                      */
/* <--> element = [a b c d e f g h i] */
/*      row_dim = 5, col_dim = 5 */
/*      nzero_index[0] = [0 2 3] */
/*      nzero_index[1] = [4]     */
/*      nzero_index[2] = [1 2]   */
/*      nzero_index[3] = [3]     */
/*      nzero_index[4] = [2 4]   */
/*      nzero_col_dim[0] = 3 */
/*      nzero_col_dim[1] = 1 */
/*      nzero_col_dim[2] = 2 */
/*      nzero_col_dim[3] = 1 */
/*      nzero_col_dim[4] = 2 */
/*      nzero_total_num = 9  */
/*      zero_element = 0.0   */

/* initialize DRSMatrix */
DRSMatrix init_drsmatrix(long int row_dim, long int *nzero_col_dim, long int nzero_total_num)
{
	DRSMatrix ret;
	long int i, j;

	ret = (DRSMatrix)malloc(sizeof(drsmatrix));
	if(ret == NULL)
	{
		fprintf(stderr, "Cannot allocate DRSMatrix\n");
		return ret;
	}
	
	if((row_dim < 0) || (nzero_total_num < 0))
	{
		fprintf(stderr, "Illigal nzero values\n");
		return NULL;
	}

	ret->row_dim = row_dim;
	ret->col_dim = row_dim; // Square matrix only
	ret->nzero_total_num = nzero_total_num;
	ret->zero_element = 0.0;

	/* allocate nzero_index */
	//printf("%d %d %d %d\n", ret->row_dim, ret->col_dim, ret->nzero_total_num, sizeof(long int *) * row_dim);
	ret->nzero_col_dim = (long int *)malloc(sizeof(long int *) * row_dim);
	ret->nzero_row_dim = (long int *)malloc(sizeof(long int *) * ret->col_dim);
	ret->nzero_index = (long int **)malloc(sizeof(long int *) * row_dim);
	if(ret->nzero_index == NULL)
	{
		fprintf(stderr, "Cannot allocate DRSMatrix(nzero_index!)\n");
		return NULL;
	}
	for(i = 0; i < row_dim; i++)
	{
		if(nzero_col_dim[i] < 0)
		{
			fprintf(stderr, "Illigal nzero values(nzero_col_dim[%d])\n", i);
			return NULL;
		}
		ret->nzero_col_dim[i] = nzero_col_dim[i];
		ret->nzero_index[i] = (long int *)malloc(sizeof(long int) * nzero_col_dim[i]);
		if(ret->nzero_index[i] == NULL)
		{
			fprintf(stderr, "Cannot allocate DRSMatrix(nzero_index[%d]!)\n", i);
			return NULL;
		}
		for(j = 0; j < nzero_col_dim[i]; j++)
			ret->nzero_index[i][j] = EMPTY;
	}

	/* allocate element */
	ret->element = (double *)malloc(sizeof(double) * nzero_total_num);
	if(ret->element == NULL)
	{
		fprintf(stderr, "Cannot allocate DRSMatrix(element!)\n");
		return NULL;
	}
	for(i = 0; i < nzero_total_num; i++)
		*(ret->element + i) = 0.0;

	return ret;
}

/* Clear DRSMatrix */
void free_drsmatrix(DRSMatrix mat)
{
	long int i;

	free(mat->element);
	free(mat->nzero_col_dim);
	free(mat->nzero_row_dim);

	for(i = 0; i < mat->row_dim; i++)
		free(mat->nzero_index[i]);

	free(mat);
}

/* set nzero_row_dim automatically */
void set_nzero_row_dim(DRSMatrix mat)
{
	long int i, j;

	/* all clear */
	for(i = 0; i < mat->col_dim; i++)
		mat->nzero_row_dim[i] = 0;

	/* search */
	for(i = 0; i < mat->row_dim; i++)
	{
		for(j = 0; j < mat->nzero_col_dim[i]; j++)
			mat->nzero_row_dim[*(mat->nzero_index[i] + j)]++;
	}
}

/* Print DRSMatrix */
void print_drsmatrix(DRSMatrix mat)
{
	long int i, j, total_index;

	if(mat == NULL)
		fprintf(stderr, "ERROR!\n");

	total_index = 0;
	for(i = 0; i < mat->row_dim; i++)
	{
		// Fix!: 2011-08-29 by T.Kouya
		if(mat->nzero_col_dim[i] >= 1)
		{
			printf("%5d: ", i);
			for(j = 0; j < mat->nzero_col_dim[i] - 1 ; j++)
				printf("%d->%e, ", mat->nzero_index[i][j], mat->element[total_index++]);
			printf("%d->%e\n", mat->nzero_index[i][mat->nzero_col_dim[i] - 1], mat->element[total_index++]);
		}
	}
	return;
}

/* initialize and substitute DRSMatrix from DMatrix */
DRSMatrix init_set_drsmatrix_dmatrix(DMatrix org_mat)
{
	long int i, j;
	long int nzero_total_num, total_index, j_index;
	long int *ptr_nzero_col_dim;
	DRSMatrix ret;

	/* initialize variables */
	nzero_total_num = 0;

	ptr_nzero_col_dim = (long int *)malloc((size_t)(sizeof(long int) * (org_mat->row_dim)));

	/* initialize num_col as numbers of nonzero element in each row */
	if(ptr_nzero_col_dim == NULL)
	{
		fprintf(stderr, "Cannot allocate num_col(size %d)\n", sizeof(long int) * (org_mat->row_dim));
		return NULL;
	}

	/* Read org_mat (1st) */
	for(i = 0; i < org_mat->row_dim; i++)
	{
		*(ptr_nzero_col_dim + i) = 0;
		for(j = 0; j < org_mat->col_dim; j++)
		{
			if(get_dmatrix_ij(org_mat, i, j) != 0.0)
			{
				/* get nzero_row_dim, nzero_col_dim, nzero_total_num */
				nzero_total_num++;
				*(ptr_nzero_col_dim + i) += 1;
			}
		}
	}

	// initialize
	ret = init_drsmatrix(org_mat->row_dim, ptr_nzero_col_dim, nzero_total_num);

	/* Read org_mat (2nd) */
	total_index = 0;
	for(i = 0; i < ret->row_dim; i++)
	{
		j_index = 0;
		for(j = 0; j < ret->col_dim; j++)
		{
			if(get_dmatrix_ij(org_mat, i, j) != 0.0)
			{
				ret->nzero_index[i][j_index] =  j;
				*(ret->element + total_index) = get_dmatrix_ij(org_mat, i, j);
				total_index += 1;
				j_index += 1;
			}
		}
	}

	free(ptr_nzero_col_dim);

	return ret;
}

/* Get variables to initialize DRSMatrix */
int get_vars_drsmatrix_fname(long int *ptr_row_dim, long int **ptr_nzero_col_dim, long int *ptr_nzero_total_num, const char *fname)
{
	FILE *fp;
	static char line_buf[LINE_BUF_LEN];
	long int i, j;
	long int row_index, col_index, max_row_index, max_col_index, nzero_total_num;
	long int ele_index;

	/* file open */
	fp = fopen(fname, "r");
	if(fp == NULL)
	{
		fprintf(stderr, "Cannot open %s\n", fname);
		return ERROR;
	}

	/* initialize variables */
	nzero_total_num = 0;
	max_row_index = 0;
	max_col_index = 0;

	/* Read file (1st) */
	while(fgets(line_buf, LINE_BUF_LEN - 1, fp) != NULL)
	{
		/* get row_index and col_index */
		sscanf(line_buf, "%ld %ld", &row_index, &col_index);

		/* get nzero_row_dim, nzero_col_dim, nzero_total_num */
		nzero_total_num++;
		//row_index--; col_index--;
		if(max_row_index < row_index)
			max_row_index = row_index;
		if(max_col_index < col_index)
			max_col_index = col_index;
	}

//	printf("Total URIs: %d\n", max_row_index + 1);

	/* initialize num_col as numbers of nonzero element in each row */
	*ptr_row_dim = max_row_index + 1;
	*ptr_nzero_total_num = nzero_total_num;
	*ptr_nzero_col_dim = (long int *)malloc((size_t)(sizeof(long int) * (max_row_index + 1)));
	if(*ptr_nzero_col_dim == NULL)
	{
		fprintf(stderr, "Cannot allocate num_col(size %d)\n", sizeof(long int) * (max_row_index + 1));
		return ERROR;
	}

	/* Read file (2nd) */
	for(i = 0; i <= max_row_index; i++)
		*(*ptr_nzero_col_dim + i) = 0;

	rewind(fp);
	while(fgets(line_buf, LINE_BUF_LEN - 1, fp) != NULL)
	{
		/* get row_index, col_index, element */
		sscanf(line_buf, "%ld %ld", &row_index, &col_index);
		//fprintf(stderr, "%ld %ld\n", row_index--, col_index--);
		*(*ptr_nzero_col_dim + row_index) += 1;
	}

	fclose(fp);

	return SUCCESS;
}

/* Read URI linking data */
int fread_urilinkdat_fname(DRSMatrix ret, const char *fname)
{
	FILE *fp;
	static char line_buf[LINE_BUF_LEN];
	static long int *nzero_row_dim;
	long int i, j;
	long int row_index, col_index, max_row_index, max_col_index, nzero_total_num;
	long int ele_index;

	/* file open */
	fp = fopen(fname, "r");
	if(fp == NULL)
	{
		fprintf(stderr, "Cannot open %s\n", fname);
		return ERROR;
	}

	nzero_row_dim = (long int *)malloc(sizeof(long int) * ret->col_dim);

	/* Read file */
	for(i = 0; i < ret->row_dim; i++)
		ret->nzero_col_dim[i] = 0;
	for(i = 0; i < ret->col_dim; i++)
		nzero_row_dim[i] = 0;
	
	while(fgets(line_buf, LINE_BUF_LEN - 1, fp) != NULL)
	//for(i = 0; i <= max_row_index; i++)
	{
		/* get row_index, col_index, element */
		sscanf(line_buf, "%ld %ld", &row_index, &col_index);
		//fscanf(fp, "%ld %ld", &row_index, &col_index);
		ret->nzero_index[row_index][ret->nzero_col_dim[row_index]] = col_index;
		//printf("%ld %ld %ld->%d\n", row_index, col_index, ret->nzero_col_dim[row_index], ret->nzero_index[row_index][ret->nzero_col_dim[row_index]]);
		ret->nzero_col_dim[row_index]++;
		nzero_row_dim[col_index]++;
	}

	ele_index = 0;
	for(i = 0; i < ret->row_dim; i++)
	{
		//printf("ret->nzero_col_dim[%d] = %d\n", i, ret->nzero_col_dim[i]);
		for(j = 0; j < ret->nzero_col_dim[i]; j++)
		{
			*(ret->element + ele_index) = 1.0 / (double)nzero_row_dim[ret->nzero_index[i][j]];
			//printf("%ld %ld %ld -> %f\n", ele_index, i, j, *(ret->element + ele_index));
			ele_index++;
		}
	}

	free(nzero_row_dim);

	/* file close */
	fclose(fp);

//	print_drsmatrix(ret);

	return SUCCESS;
}

/* Multiply DRSMatrix * DVector */
int mul_drsmatrix_dvec(DVector ret, DRSMatrix mat, DVector vec)
{
	long int i, j, total_index;

	if((ret->dim < mat->col_dim) || (vec->dim != mat->col_dim))
	{
		fprintf(stderr, "Illegal dimension!\n");
		return ERROR;
	}

	total_index = 0;
	for(i = 0; i < mat->row_dim; i++)
	{
		ret->element[i] = 0.0;
		for(j = 0; j < mat->nzero_col_dim[i]; j++)
		{
// fix!: 2011-08-29 by T.Kouya
			ret->element[i] += mat->element[total_index] * vec->element[mat->nzero_index[i][j]];
			total_index++;
		}
	}

	return SUCCESS;
}

/* Multiply DRSMatrix^T * DVector */
int mul_drsmatrixt_dvec(DVector ret, DRSMatrix mat, DVector vec)
{
	long int i, j, total_index;

	if((ret->dim < mat->col_dim) || (vec->dim != mat->col_dim))
	{
		fprintf(stderr, "Illegal dimension!\n");
		return ERROR;
	}

	for(i = 0; i < mat->row_dim; i++)
		ret->element[i] = 0.0;

	total_index = 0;
	for(i = 0; i < mat->row_dim; i++)
	{
		for(j = 0; j < mat->nzero_col_dim[i]; j++)
		{
			ret->element[mat->nzero_index[i][j]] += mat->element[total_index] * vec->element[i];
			total_index++;
		}
	}

	return SUCCESS;
}

/* Power Method for Randomly Sparse Matrices */
/* 	double *evec: the eigenvector for max eigenvalue */
/* 	double *drsmat: Randomly sparse matrix */
/* 	double reps, aeps: Relative and Absolute tolerance */
/* 	long int max_times: Maximum iterative times of Power method */
double dpower_rsmatrix(DVector evec, DRSMatrix mat, double reps, double aeps, long int max_times)
{
	long int i, absmax_index, times;
	double absmax_new_evec, max_eig, old_max_eig;
	DVector new_evec;

	new_evec = init_dvector(mat->row_dim);

	/* initialize evec */
	for(i = 0; i < evec->dim; i++)
		evec->element[i] = 1.0;

	/* main loop */
	old_max_eig = 0.0;
	for(times = 0; times < max_times; times++)
	{
		/* w := A * x */
		mul_drsmatrix_dvec(new_evec, mat, evec);
		absmax_index = absmax_index_dvector(&absmax_new_evec, new_evec);
		max_eig = absmax_new_evec / evec->element[absmax_index];
//		smul_dvector(evec, 1.0 / absmax_new_evec, new_evec);
		smul_dvector(evec, 1.0 / norm1_dvector(new_evec), new_evec); // Baba's example
		if((fabs(max_eig - old_max_eig) <= reps * fabs(old_max_eig) + aeps) && (times >= 2))
		{
			fprintf(stderr, "Convergent!(Iterative Times = %d)\n", times);
			break;
		}
		if(times % 10 == 0)
			fprintf(stderr, "%5d %25.17e\n", times, max_eig);
		old_max_eig = max_eig;
	}

	return max_eig;
}

/* Scalar multiply of DVector */
int smul_dvector(DVector ret, double scalar, DVector vec)
{
	long int i;

	if(ret->dim < vec->dim)
		return ERROR;

	for(i = 0; i < vec->dim; i++)
		ret->element[i] = scalar * vec->element[i];

	return SUCCESS;
}

/* Select index of absolute maximum element and its value in DVector */
long int absmax_index_dvector(double *ret, DVector vec)
{
	long int absmax_index, i;
	double abs_element;

	*ret = 0.0;
	absmax_index = 0;
	for(i = 0; i < vec->dim; i++)
	{
		abs_element = fabs(vec->element[i]);
		if(*ret < abs_element)
		{
			absmax_index = i;
			//*ret = abs_element;
		}
	}

	*ret = vec->element[absmax_index];

	return absmax_index;
}
