/********************************************************************************/
/*                                                                              */
/* sparse_mpf.c : Sparse Matrix and Vector Library (Multiple Precision)         */
/* Copyright (c) 2011 Tomonori Kouya, All rights reserved.                      */
/*                                                                              */
/* Version 0.1 2011-08-29 : Bug fix in mul_mpfrsmatrix_mpfvec                   */
/* Version 0.1 2012-03-23 : Bug fix in init2_mpfrsmatrix                        */
/*                                                                              */
/* This program is free software: you can redistribute it and/or modify it      */
/* under the terms of the GNU Lesser General Public License as published by the */
/* Free Software Foundation, either version 3 of the License or any later       */
/* version.                                                                     */
/*                                                                              */
/* This program is distributed in the hope that it will be useful, but WITHOUT  */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or        */
/* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License */
/* for more details.                                                            */
/*                                                                              */
/* You should have received a copy of the GNU Lesser General Public License     */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                              */
/********************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "bncsparse.h"

#ifdef USE_GMP

/* Sparse matrix struct */
/* Example:             */
/*      0 1 2 3 4       */
/* A = [a 0 b c 0]0     */
/*     [0 d 0 0 0]1     */
/*     [0 e f 0 0]2     */
/*     [0 0 0 g 0]3     */
/*     [0 0 h 0 i]4     */
/*                      */
/* <--> element = [a b c d e f g h i] */
/*      row_dim = 5, col_dim = 5 */
/*      nzero_index[0] = [0 2 3] */
/*      nzero_index[1] = [4]     */
/*      nzero_index[2] = [1 2]   */
/*      nzero_index[3] = [3]     */
/*      nzero_index[4] = [2 4]   */
/*      nzero_col_dim[0] = 3 */
/*      nzero_col_dim[1] = 1 */
/*      nzero_col_dim[2] = 2 */
/*      nzero_col_dim[3] = 1 */
/*      nzero_col_dim[4] = 2 */
/*      nzero_total_num = 9  */
/*	    zero_element = 0UL   */

/* initialize MPFRSMatrix */
MPFRSMatrix init_mpfrsmatrix(long int row_dim, long int *nzero_col_dim, long int nzero_total_num)
{
	MPFRSMatrix ret;
	long int i, j;

	ret = (MPFRSMatrix)malloc(sizeof(mpfrsmatrix));
	if(ret == NULL)
	{
		fprintf(stderr, "Cannot allocate MPFRSMatrix\n");
		return ret;
	}
	
	if((row_dim < 0) || (nzero_total_num < 0))
	{
		fprintf(stderr, "Illigal nzero values\n");
		return NULL;
	}

	ret->row_dim = row_dim;
	ret->col_dim = row_dim; // Square matrix only
	ret->nzero_total_num = nzero_total_num;
	ret->prec = get_bnc_default_prec();
	mpf_init2(ret->zero_element, ret->prec);
	mpf_set_ui(ret->zero_element, 0UL);

	/* allocate nzero_index */
	//printf("%d %d %d %d\n", ret->row_dim, ret->col_dim, ret->nzero_total_num, sizeof(long int *) * row_dim);
	ret->nzero_col_dim = (long int *)malloc(sizeof(long int *) * row_dim);
	ret->nzero_row_dim = (long int *)malloc(sizeof(long int *) * ret->col_dim);
	ret->nzero_index = (long int **)malloc(sizeof(long int *) * row_dim);
	if(ret->nzero_index == NULL)
	{
		fprintf(stderr, "Cannot allocate MPFRSMatrix(nzero_index!)\n");
		return NULL;
	}
	for(i = 0; i < row_dim; i++)
	{
		if(nzero_col_dim[i] < 0)
		{
			fprintf(stderr, "Illigal nzero values(nzero_col_dim[%d])\n", i);
			return NULL;
		}
		ret->nzero_col_dim[i] = nzero_col_dim[i];
		ret->nzero_index[i] = (long int *)malloc(sizeof(long int) * nzero_col_dim[i]);
		if(ret->nzero_index[i] == NULL)
		{
			fprintf(stderr, "Cannot allocate MPFRSMatrix(nzero_index[%d]!)\n", i);
			return NULL;
		}
		for(j = 0; j < nzero_col_dim[i]; j++)
			ret->nzero_index[i][j] = EMPTY;
	}
	
	/* allocate element */
	//ret->element = (mpf_ptr)malloc(sizeof(mpf_t) * nzero_total_num);
	ret->element = (mpf_t *)calloc(sizeof(mpf_t), nzero_total_num);
	if(ret->element == NULL)
	{
		fprintf(stderr, "Cannot allocate MPFRSMatrix(element!)\n");
		return NULL;
	}
//	for(i = 0; i < nzero_total_num; i++)
//		*(ret->element + i) = 0.0;
	for(i = 0; i < nzero_total_num; i++)
	{
		mpf_init(*(ret->element + i));
		mpf_set_ui(*(ret->element + i), 0UL);
//		mpf_init_set_ui(*(ret->element + i), 0UL);
	}

	return ret;
}

/* initialize MPFRSMatrix with specific precision */
MPFRSMatrix init2_mpfrsmatrix(long int row_dim, long int *nzero_col_dim, long int nzero_total_num, unsigned long prec)
{
	MPFRSMatrix ret;
	long int i, j;

	ret = (MPFRSMatrix)malloc(sizeof(mpfrsmatrix));
	if(ret == NULL)
	{
		fprintf(stderr, "Cannot allocate MPFRSMatrix\n");
		return ret;
	}
	
	if((row_dim < 0) || (nzero_total_num < 0))
	{
		fprintf(stderr, "Illigal nzero values\n");
		return NULL;
	}

	ret->row_dim = row_dim;
	ret->col_dim = row_dim; // Square matrix only
	ret->nzero_total_num = nzero_total_num;
	ret->prec = prec; // Fix!
	mpf_init2(ret->zero_element, ret->prec);
	mpf_set_ui(ret->zero_element, 0UL);

	/* allocate nzero_index */
	//printf("%d %d %d %d\n", ret->row_dim, ret->col_dim, ret->nzero_total_num, sizeof(long int *) * row_dim);
	ret->nzero_col_dim = (long int *)malloc(sizeof(long int *) * row_dim);
	ret->nzero_row_dim = (long int *)malloc(sizeof(long int *) * ret->col_dim);
	ret->nzero_index = (long int **)malloc(sizeof(long int *) * row_dim);
	if(ret->nzero_index == NULL)
	{
		fprintf(stderr, "Cannot allocate MPFRSMatrix(nzero_index!)\n");
		return NULL;
	}
	for(i = 0; i < row_dim; i++)
	{
		if(nzero_col_dim[i] < 0)
		{
			fprintf(stderr, "Illigal nzero values(nzero_col_dim[%d])\n", i);
			return NULL;
		}
		ret->nzero_col_dim[i] = nzero_col_dim[i];
		ret->nzero_index[i] = (long int *)malloc(sizeof(long int) * nzero_col_dim[i]);
		if(ret->nzero_index[i] == NULL)
		{
			fprintf(stderr, "Cannot allocate MPFRSMatrix(nzero_index[%d]!)\n", i);
			return NULL;
		}
		for(j = 0; j < nzero_col_dim[i]; j++)
			ret->nzero_index[i][j] = EMPTY;
	}
	
	/* allocate element */
	//ret->element = (mpf_ptr)malloc(sizeof(mpf_t) * nzero_total_num);
	ret->element = (mpf_t *)calloc(sizeof(mpf_t), nzero_total_num);
	if(ret->element == NULL)
	{
		fprintf(stderr, "Cannot allocate MPFRSMatrix(element!)\n");
		return NULL;
	}
//	for(i = 0; i < nzero_total_num; i++)
//		*(ret->element + i) = 0.0;
	for(i = 0; i < nzero_total_num; i++)
	{
		mpf_init2(*(ret->element + i), prec);
		mpf_set_ui(*(ret->element + i), 0UL);
//		mpf_init_set_ui(*(ret->element + i), 0UL);
	}

	return ret;
}

/* Clear MPFRSMatrix */
void free_mpfrsmatrix(MPFRSMatrix mat)
{
	long int i;

//	free(mat->element);
	free(mat->nzero_col_dim);
	free(mat->nzero_row_dim);
	mpf_clear(mat->zero_element);

	for(i = 0; i < mat->row_dim; i++)
		free(mat->nzero_index[i]);

	//free(mat);
	if(mat->element != NULL)
		for(i = 0; i < mat->nzero_total_num; i++)
				mpf_clear(*(mat->element + i));

	free(mat);
}

/* set nzero_row_dim automatically */
void set_nzero_row_dim_mpf(MPFRSMatrix mat)
{
	long int i, j;

	/* all clear */
	for(i = 0; i < mat->col_dim; i++)
		mat->nzero_row_dim[i] = 0;

	/* search */
	for(i = 0; i < mat->row_dim; i++)
	{
		for(j = 0; j < mat->nzero_col_dim[i]; j++)
			mat->nzero_row_dim[*(mat->nzero_index[i] + j)]++;
	}
}

/* Print MPFRSMatrix */
void print_mpfrsmatrix(MPFRSMatrix mat)
{
	long int i, j, total_index;

	if(mat == NULL)
		fprintf(stderr, "ERROR!\n");

	total_index = 0;
	for(i = 0; i < mat->row_dim; i++)
	{
		// Fix!: 2011-08-29 by T.Kouya
		if(mat->nzero_col_dim[i] >= 1)
		{
			printf("%5d: ", i);
			for(j = 0; j < mat->nzero_col_dim[i] - 1 ; j++)
			{
				// printf("%d->%f, ", mat->nzero_index[i][j], mat->element[total_index++]);
				printf("%d-> ", mat->nzero_index[i][j]);
				mpf_out_str(stdout, 10, 0, mat->element[total_index++]);
				printf(", ");
			}
			//printf("%d->%f\n", mat->nzero_index[i][mat->nzero_col_dim[i] - 1], mat->element[total_index++]);
			printf("%d->", mat->nzero_index[i][mat->nzero_col_dim[i] - 1]);
			mpf_out_str(stdout, 10, 0, mat->element[total_index++]);
			printf("\n");
		}
	}
	return;
}

/* initialize and substitute MPFRSMatrix from MPFMatrix */
MPFRSMatrix init_set_mpfrsmatrix_mpfmatrix(MPFMatrix org_mat)
{
	long int i, j;
	long int nzero_total_num, total_index, j_index;
	long int *ptr_nzero_col_dim;
	MPFRSMatrix ret;

	/* initialize variables */
	nzero_total_num = 0;

	ptr_nzero_col_dim = (long int *)malloc((size_t)(sizeof(long int) * (org_mat->row_dim)));

	/* initialize num_col as numbers of nonzero element in each row */
	if(ptr_nzero_col_dim == NULL)
	{
		fprintf(stderr, "Cannot allocate num_col(size %d)\n", sizeof(long int) * (org_mat->row_dim));
		return NULL;
	}

	/* Read org_mat (1st) */
	for(i = 0; i < org_mat->row_dim; i++)
	{
		*(ptr_nzero_col_dim + i) = 0;
		for(j = 0; j < org_mat->col_dim; j++)
		{
			if(mpf_cmp_ui(get_mpfmatrix_ij(org_mat, i, j), 0UL) != 0)
			{
				/* get nzero_row_dim, nzero_col_dim, nzero_total_num */
				nzero_total_num++;
				*(ptr_nzero_col_dim + i) += 1;
			}
		}
	}

	// initialize
	ret = init2_mpfrsmatrix(org_mat->row_dim, ptr_nzero_col_dim, nzero_total_num, org_mat->prec);

	/* Read org_mat (2nd) */
	total_index = 0;
	for(i = 0; i < ret->row_dim; i++)
	{
		j_index = 0;
		for(j = 0; j < ret->col_dim; j++)
		{
			if(mpf_cmp_ui(get_mpfmatrix_ij(org_mat, i, j), 0UL) != 0)
			{
				ret->nzero_index[i][j_index] =  j;
				mpf_set(*(ret->element + total_index), get_mpfmatrix_ij(org_mat, i, j));
				total_index += 1;
				j_index += 1;
			}
		}
	}

	free(ptr_nzero_col_dim);

	return ret;
}

/* Get variables to initialize MPFRSMatrix */
int get_vars_mpfrsmatrix_fname(long int *ptr_row_dim, long int **ptr_nzero_col_dim, long int *ptr_nzero_total_num, const char *fname)
{
	FILE *fp;
	static char line_buf[LINE_BUF_LEN];
	long int i, j;
	long int row_index, col_index, max_row_index, max_col_index, nzero_total_num;
	long int ele_index;

	/* file open */
	fp = fopen(fname, "r");
	if(fp == NULL)
	{
		fprintf(stderr, "Cannot open %s\n", fname);
		return ERROR;
	}

	/* initialize variables */
	nzero_total_num = 0;
	max_row_index = 0;
	max_col_index = 0;

	/* Read file (1st) */
	while(fgets(line_buf, LINE_BUF_LEN - 1, fp) != NULL)
	{
		/* get row_index and col_index */
		sscanf(line_buf, "%ld %ld", &row_index, &col_index);

		/* get nzero_row_dim, nzero_col_dim, nzero_total_num */
		nzero_total_num++;
		//row_index--; col_index--;
		if(max_row_index < row_index)
			max_row_index = row_index;
		if(max_col_index < col_index)
			max_col_index = col_index;
	}

//	printf("Total URIs: %d\n", max_row_index + 1);

	/* initialize num_col as numbers of nonzero element in each row */
	*ptr_row_dim = max_row_index + 1;
	*ptr_nzero_total_num = nzero_total_num;
	*ptr_nzero_col_dim = (long int *)malloc((size_t)(sizeof(long int) * (max_row_index + 1)));
	if(*ptr_nzero_col_dim == NULL)
	{
		fprintf(stderr, "Cannot allocate num_col(size %d)\n", sizeof(long int) * (max_row_index + 1));
		return ERROR;
	}

	/* Read file (2nd) */
	for(i = 0; i <= max_row_index; i++)
		*(*ptr_nzero_col_dim + i) = 0;

	rewind(fp);
	while(fgets(line_buf, LINE_BUF_LEN - 1, fp) != NULL)
	{
		/* get row_index, col_index, element */
		sscanf(line_buf, "%ld %ld", &row_index, &col_index);
		//fprintf(stderr, "%ld %ld\n", row_index--, col_index--);
		*(*ptr_nzero_col_dim + row_index) += 1;
	}

	fclose(fp);

	return SUCCESS;
}

/* Read URI linking data */
int fread_urilinkdat_fname_mpf(MPFRSMatrix ret, const char *fname)
{
	FILE *fp;
	static char line_buf[LINE_BUF_LEN];
	static long int *nzero_row_dim;
	long int i, j;
	long int row_index, col_index, max_row_index, max_col_index, nzero_total_num;
	long int ele_index;
	unsigned long prec;
	mpf_t tmp;

	// init tmp
	prec = ret->prec;
	mpf_init2(tmp, prec);

	/* file open */
	fp = fopen(fname, "r");
	if(fp == NULL)
	{
		fprintf(stderr, "Cannot open %s\n", fname);
		return ERROR;
	}

	nzero_row_dim = (long int *)malloc(sizeof(long int) * ret->col_dim);

	/* Read file */
	for(i = 0; i < ret->row_dim; i++)
		ret->nzero_col_dim[i] = 0;
	for(i = 0; i < ret->col_dim; i++)
		nzero_row_dim[i] = 0;
	
	while(fgets(line_buf, LINE_BUF_LEN - 1, fp) != NULL)
	//for(i = 0; i <= max_row_index; i++)
	{
		/* get row_index, col_index, element */
		sscanf(line_buf, "%ld %ld", &row_index, &col_index);
		//fscanf(fp, "%ld %ld", &row_index, &col_index);
		ret->nzero_index[row_index][ret->nzero_col_dim[row_index]] = col_index;
		//printf("%ld %ld %ld->%d\n", row_index, col_index, ret->nzero_col_dim[row_index], ret->nzero_index[row_index][ret->nzero_col_dim[row_index]]);
		ret->nzero_col_dim[row_index]++;
		nzero_row_dim[col_index]++;
	}

	ele_index = 0;
	for(i = 0; i < ret->row_dim; i++)
	{
		//printf("ret->nzero_col_dim[%d] = %d\n", i, ret->nzero_col_dim[i]);
		for(j = 0; j < ret->nzero_col_dim[i]; j++)
		{
			//*(ret->element + ele_index) = 1.0 / (mpf_t)nzero_row_dim[ret->nzero_index[i][j]];
			mpf_set_ui(tmp, (unsigned long)nzero_row_dim[ret->nzero_index[i][j]]);
			mpf_ui_div(*(ret->element + ele_index), 1UL, tmp);

			//printf("%ld %ld %ld -> %f\n", ele_index, i, j, *(ret->element + ele_index));
			ele_index++;
		}
	}

	mpf_clear(tmp);
	free(nzero_row_dim);

	/* file close */
	fclose(fp);

//	print_MPFRSMatrix(ret);

	return SUCCESS;
}


/* Multiply MPFRSMatrix * MPFVector */
int mul_mpfrsmatrix_mpfvec(MPFVector ret, MPFRSMatrix mat, MPFVector vec)
{
	long int i, j, total_index;
	unsigned long prec;
	mpf_t tmp;

	if((ret->dim < mat->col_dim) || (vec->dim != mat->col_dim))
	{
		fprintf(stderr, "Illegal dimension!\n");
		return ERROR;
	}

	prec = ret->prec;
	mpf_init2(tmp, prec);

	total_index = 0;
	for(i = 0; i < mat->row_dim; i++)
	{
		//ret->element[i] = 0.0;
		mpf_set_ui(ret->element[i], 0UL);
		for(j = 0; j < mat->nzero_col_dim[i]; j++)
		{
			//ret->element[i] += mat->element[total_index] * vec->element[mat->nzero_index[i][j]];
			mpf_mul(tmp, mat->element[total_index], vec->element[mat->nzero_index[i][j]]);
			mpf_add(ret->element[i], ret->element[i], tmp);
			total_index++;
		}
	}

	mpf_clear(tmp);

	return SUCCESS;
}
/* Multiply MPFRSMatrix^T * MPFVector */
int mul_mpfrsmatrixt_mpfvec(MPFVector ret, MPFRSMatrix mat, MPFVector vec)
{
	long int i, j, total_index;
	unsigned long prec;
	mpf_t tmp;

	if((ret->dim < mat->col_dim) || (vec->dim != mat->col_dim))
	{
		fprintf(stderr, "Illegal dimension!\n");
		return ERROR;
	}

	prec = ret->prec;
	mpf_init2(tmp, prec);

	for(i = 0; i < mat->row_dim; i++)
		mpf_set_ui(ret->element[i], 0UL);

	total_index = 0;
	for(i = 0; i < mat->row_dim; i++)
	{
		for(j = 0; j < mat->nzero_col_dim[i]; j++)
		{
			//ret->element[mat->nzero_index[i][j]] += mat->element[total_index] * vec->element[i];
			mpf_mul(tmp, mat->element[total_index], vec->element[i]);
			mpf_add(ret->element[mat->nzero_index[i][j]], ret->element[mat->nzero_index[i][j]], tmp);
			total_index++;
		}
	}

	mpf_clear(tmp);

	return SUCCESS;
}

/* Power Method for Randomly Sparse Matrices */
/* 	mpf_t *evec: the eigenvector for max eigenvalue */
/* 	mpf_t *drsmat: Randomly sparse matrix */
/* 	mpf_t reps, aeps: Relative and Absolute tolerance */
/* 	long int max_times: Maximum iterative times of Power method */
void mpfpower_rsmatrix(mpf_t max_eig, MPFVector evec, MPFRSMatrix mat, mpf_t reps, mpf_t aeps, long int max_times)
{
	long int i, absmax_index, times;
	mpf_t absmax_new_evec, old_max_eig, tmp, tmp2;
	MPFVector new_evec;
	unsigned long prec;

	prec = evec->prec;

	mpf_init2(absmax_new_evec, prec);
//	mpf_init2(max_eig, prec);
	mpf_init2(old_max_eig, prec);
	mpf_init2(tmp, prec);
	mpf_init2(tmp2, prec);

	new_evec = init2_mpfvector(mat->row_dim, prec);

	/* initialize evec */
	for(i = 0; i < evec->dim; i++)
		mpf_set_ui(evec->element[i], 1UL);
		//evec->element[i] = 1.0;

	/* main loop */
	//old_max_eig = 0.0;
	mpf_set_ui(old_max_eig, 0UL);
	for(times = 0; times < max_times; times++)
	{
		/* w := A * x */
		mul_mpfrsmatrix_mpfvec(new_evec, mat, evec);
		absmax_index = absmax_index_mpfvector(absmax_new_evec, new_evec);
//		max_eig = absmax_new_evec / evec->element[absmax_index];
		mpf_div(max_eig, absmax_new_evec, evec->element[absmax_index]);
//		smul_mpfvector(evec, 1.0 / absmax_new_evec, new_evec);
//		smul_mpfvector(evec, 1.0 / norm1_MPFVector(new_evec), new_evec); // Baba's example
		norm1_mpfvector(tmp, new_evec);
		mpf_ui_div(tmp, 1UL, tmp);
		smul_mpfvector(evec, tmp, new_evec);

//		if((fabs(max_eig - old_max_eig) <= reps * fabs(old_max_eig) + aeps) && (times >= 2))
		mpf_sub(tmp, max_eig, old_max_eig);
		mpf_abs(tmp, tmp);
		
		mpf_abs(tmp2, old_max_eig);
		mpf_mul(tmp2, reps, tmp2);
		mpf_add(tmp2, tmp2, aeps);
		if((mpf_cmp(tmp, tmp2) <= 0) && (times >= 2))
		{
			fprintf(stderr, "Convergent!(Iterative Times = %d)\n", times);
			break;
		}
		if(times % 10 == 0)
			fprintf(stderr, "%5d %25.17e\n", times, max_eig);
		//old_max_eig = max_eig;
		mpf_set(old_max_eig, max_eig);
	}

//	mpf_clear(max_eig);
	mpf_clear(old_max_eig);
	mpf_clear(absmax_new_evec);
	mpf_clear(tmp);
	mpf_clear(tmp2);

	free_mpfvector(new_evec);

//	return max_eig;
	return;
}

/* Scalar multiply of MPFVector */
int smul_mpfvector(MPFVector ret, mpf_t scalar, MPFVector vec)
{
	long int i;

	if(ret->dim < vec->dim)
		return ERROR;

	for(i = 0; i < vec->dim; i++)
		mpf_mul(ret->element[i],  scalar, vec->element[i]);

	return SUCCESS;
}

/* Select index of absolute maximum element and its value in MPFVector */
long int absmax_index_mpfvector(mpf_t ret, MPFVector vec)
{
	long int absmax_index, i;
	mpf_t abs_element;

	mpf_init2(abs_element, mpf_get_prec(ret));

	mpf_set_ui(ret, 0UL);
	absmax_index = 0;
	for(i = 0; i < vec->dim; i++)
	{
		mpf_abs(abs_element, vec->element[i]);
		if(mpf_cmp(ret, abs_element) < 0)
		{
			absmax_index = i;
			//*ret = abs_element;
		}
	}

	mpf_set(ret, vec->element[absmax_index]);

	mpf_clear(abs_element);

	return absmax_index;
}

#endif
